'''
A collection of analysis functions that are somewhat connected to halo properties.

Examples:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_320', physical=True)
    >>> center = shrinking_sphere(s.stars, center=[s.boxsize/2]*3,
    ...                           R=s.boxsize*np.sqrt(3)) # doctest: +ELLIPSIS
    load block pos... done.
    do a shrinking sphere...
      starting values:
        center = ...
        R      = ...
    load block mass... done.
    done.
    >>> if np.linalg.norm( center - UnitArr([33816.9, 34601.1, 32681.0], 'kpc') ) > 1.0:
    ...     print(center)

    >>> R200, M200 = virial_info(s, center)
    >>> if abs(R200 - '177 kpc') > 3 or abs(M200 - '1e12 Msol') / '1e12 Msol' > 0.1:
    ...     print(R200, M200)
    >>> Translation(-center).apply(s)
    apply Translation to "pos" of "snap_M1196_4x_320"... done.

    >>> l3 = 1500.*s.cosmology.rho_crit(s.redshift)/np.median(s.highres.dm['mass'])
    >>> FoF, N_FoF = find_FoF_groups(s.highres.dm,
    ...             l=l3**Fraction(-1,3),
    ...             dvmax='100 km/s') # doctest: +ELLIPSIS
    load block vel... done.
    perform a FoF search on 1,001,472 particles:
      l      = 2.2 [kpc]
      dv_max = 1e+02 [km s**-1]
      N     >= 100
    found 90 groups
    the 3 most massive ones are:
      group 0:   1.02e+11 [Msol]  @  [0.333, 0.255, 0.111] [kpc]
      group 1:    4.1e+10 [Msol]  @  [492, 941, 429] [kpc]
      group 2:   8.37e+09 [Msol]  @  [-1.57e+03, -1.4e+03, -1.11e+03] [kpc]

    # find galaxies (exclude those with almost only gas)
    >>> galaxies = generate_FoF_catalogue(s.baryons,
    ...             min_N=300,
    ...             dvmax='100 km/s', #max_halos=5,
    ...             progressbar=False,
    ...             exclude=lambda g,s: g.Mgas/g.mass>0.9)  # doctest: +ELLIPSIS
    perform a FoF search on 1,001,472 particles:
      l      = 1.3 [kpc]
      dv_max = 1e+02 [km s**-1]
      N     >= 300
    found 4 groups
    the 3 most massive ones are:
      group 0:   3.12e+10 [Msol]  @  [0.0693, 0.218, 0.0566] [kpc]
      group 1:   6.94e+09 [Msol]  @  [492, 940, 428] [kpc]
      group 2:   3.36e+09 [Msol]  @  [871, 1.36e+03, 729] [kpc]
    initialize halos from FoF group IDs...
    load block ID... done.
    load block Z... done.
    derive block elements... done.
    derive block H... done.
    load block ne... done.
    load block u... done.
    derive block temp... done.
    load block rho... done.
    load block form_time... done.
    derive block age... done.
    initialized 3 halos.
    >>> galaxies[0] # doctest: +ELLIPSIS
    <Halo @0x..., N = 55,..., M = 3.1e+10 [Msol]>
    >>> gal = s[galaxies[0]]
    >>> assert len(gal) == len(galaxies[0])
    >>> assert set(gal['ID']) == set(galaxies[0].IDs)
    >>> assert np.all(gal.parts == np.array(galaxies[0].parts))
    >>> assert gal['mass'].sum() == galaxies[0].props['mass']
    >>> assert gal.stars['mass'].sum() == galaxies[0].Mstars
    >>> gal = galaxies[0]
    >>> gal.calc_prop('ssc', root=s)    # doctest:+ELLIPSIS
    UnitArr([...], units="kpc")
    >>> if np.linalg.norm(gal.ssc - gal.com) > 1000:
    ...     print(gal.ssc)
    ...     print(gal.com)
    >>> assert 'Rmax' in Halo.calculable_props()
    >>> Halo.prop_descr('M200_com')
    'spherical M200 with `virial_info` with com as center'

    Test pickling:
    >>> import pickle as pickle
    >>> pkld_gal = pickle.dumps(gal)
    >>> gal_2 = pickle.loads(pkld_gal)
    >>> assert np.all(gal_2.IDs == gal.IDs)
    >>> assert set(gal_2.props.keys()) == set(gal.props.keys())
    >>> assert np.all(gal_2.com == gal.com)
'''
__all__ = ['shrinking_sphere', 'virial_info', 'find_FoF_groups',
           'NO_FOF_GROUP_ID', 'Rockstar_halo_field_names',
           'Rockstar_particle_field_names', 'RockstarHeader',
           'read_Rockstar_file', 'generate_Rockstar_halos', 'Halo',
           'nxt_ngb_dist_perc', 'generate_FoF_catalogue',
           'find_most_massive_progenitor']

import numpy as np
from .. import utils
from ..units import *
from ..utils import *
import sys, os
from ..transformation import *
from .properties import *
from ..snapshot import *
from .. import environment
from .. import C


def shrinking_sphere(s, center, R, periodic=True, shrink_factor=0.93,
                     stop_N=10, verbose=None):
    '''
    Find the densest point by shrinking sphere technique.

    Technique is described in Power et al. (2003).

    Args:
        s (Snap):               The (sub-)snapshot to find the densest point for.
        center (array-like):    The center to start with.
        R (float, UnitArr, str):The initial radius.
        periodic (bool):        Whether to assume a periodic box (with the
                                sidelength of snap.props['boxsize']) or not.
        shrink_factor (float):  The factor to shrink the sphere in each step.
        stop_N (int):           If so many or less particles are left in the
                                sphere, stop.
        verbose (int):          Verbosity level. Default: the gobal pygad
                                verbosity level.

    Returns:
        center (UnitArr):       The center.
    '''
    if verbose is None:
        verbose = environment.verbose
    center0 = UnitQty(center, s['pos'].units, subs=s, dtype=np.float64)
    R = UnitScalar(R, s['pos'].units, subs=s)

    if verbose >= environment.VERBOSE_NORMAL:
        print('do a shrinking sphere...')
        print('  starting values:')
        print('    center = %s' % center0)
        print('    R      = %s' % R)
        sys.stdout.flush()

    if not 0 < shrink_factor < 1:
        raise ValueError('"shrink_factor" must be in the interval (0,1)!')
    if not 0 < stop_N:
        raise ValueError('"stop_N" must be positive!')

    pos = s['pos'].astype(np.float64)
    mass = s['mass'].astype(np.float64)
    assert len(pos) == len(mass)
    boxsize = float(s.boxsize.in_units_of(s['pos'].units))

    # needed since C does not know about stridings
    if pos.base is not None:
        pos = pos.copy()
        mass = mass.copy()

    center = np.empty((3,), dtype=np.float64)
    if periodic:
        C.cpygad.shrinking_sphere_periodic(
            C.c_void_p(center.ctypes.data),
            C.c_size_t(len(pos)),
            C.c_void_p(pos.ctypes.data),
            C.c_void_p(mass.ctypes.data),
            C.c_void_p(center0.ctypes.data), C.c_double(R),
            C.c_double(shrink_factor), C.c_size_t(stop_N),
            C.c_double(boxsize))
    else:
        C.cpygad.shrinking_sphere_nonperiodic(
            C.c_void_p(center.ctypes.data),
            C.c_size_t(len(pos)),
            C.c_void_p(pos.ctypes.data),
            C.c_void_p(mass.ctypes.data),
            C.c_void_p(center0.ctypes.data), C.c_double(R),
            C.c_double(shrink_factor), C.c_size_t(stop_N))
    center = center.view(UnitArr)
    center.units = s['pos'].units

    if verbose >= environment.VERBOSE_NORMAL:
        print('done.')
        sys.stdout.flush()

    return center


def virial_info(s, center=None, odens=200.0, N_min=10):
    '''
    Return the virial radius (R_odens) and the virial mass (M_odens) of the
    structure at 'center'.

    The virial radius is the radius, where the average density within that radius
    is <odens> times the critical density of the universe. If the inner <N_min>
    particles do not exceed a density of <odens> times the critical density,
    R_<odens> and M_<odens> are both returned as zero.

    Args:
        s (Snap):               The (sub-)snapshot to use.
        center (array-like):    The center of the structure to calculate the
                                properties for. (default: (0,0,0))
        odens (float):          Overdensity parameter (rho/rho_crit) used as
                                threshold for calculated radius and enclosed
                                mass to be returned.
        N_min (int):            The minimum number of particles required to form
                                the halo (have a density higher than
                                odens * rho_crit). If not reached, the return
                                values are NaN.

    Returns:
        R_odens, M_odens (both UnitArr):
                                The virial radius & mass (or corresponding
                                R_odens, M_odens)
    '''
    if center is None:
        center = [0, 0, 0]
    center = UnitQty(center, s['pos'].units, subs=s).view(np.ndarray)

    rho_crit = s.cosmology.rho_crit(z=s.redshift)
    rho_crit = rho_crit.in_units_of(s['mass'].units / s['pos'].units ** 3, subs=s)

    # make use of the potentially precalculated derived block r
    if np.all(center == 0):
        r = s['r']
    else:
        r = dist(s['pos'], center)

    mass = s['mass'].astype(np.float64)
    r = r.astype(np.float64)
    if mass.base is not None:
        mass = mass.copy()
        r = r.copy()
    info = np.empty((2,), dtype=np.float64)
    C.cpygad.virial_info(C.c_size_t(len(r)),
                         C.c_void_p(mass.ctypes.data),
                         C.c_void_p(r.ctypes.data),
                         C.c_double(odens * rho_crit),
                         C.c_size_t(N_min),
                         C.c_void_p(info.ctypes.data),
                         )
    if info[0] == 0.0:
        info[:] = np.nan
    return UnitArr(info[0], s['pos'].units), \
           UnitArr(info[1], s['mass'].units)


NO_FOF_GROUP_ID = int(np.array(-1, np.uintp))


def find_FoF_groups(s, l, dvmax=np.inf, min_N=100, sort=True, periodic_boundary=2, boxsize_manual=None, verbose=None):
    '''
    Perform a friends-of-friends search on a (sub-)snapshot.

    Args:
        s (Snap):           The (sub-)snapshot to perform the FoF finder on.
        l (UnitScalar):     The linking length to use for the FoF finder (for dark
                            matter / all matter it turned out that
                            ( 1500. * rho_crit / median(mass) )^(-1/3) is a
                            reasonable value).
        dvmax (UnitScalar): The "linking velocity" to use for the FoF finder.
        min_N (int):        The minimum number of particles in a FoF group to
                            actually define it as such.
        sort (bool):        Whether to sort the groups by mass. If True, the group
                            with ID 0 will be the most massive one and the in
                            descending order.
        periodic_boundary (int): Whether the simulation uses periodic boundary
                                 conditions or not:
                                     - 0: no periodic boundary
                                     - 1: periodic boundary turned on
                                     - 2: automatic determination, turned on for s.cosmological == True
                                     - 3: periodic boundary turned on with boxsize from boxsize_manual argument
        boxsize_manual (UnitScalar) : Boxsize for periodic boundary conditions.
        verbose (int):      Verbosity level. Default: the gobal pygad verbosity
                            level.

    Returns:
        FoF (np.ndarray):   A block of FoF group IDs for the particles of `s`.
                            (IDs are ordered in mass, if `sort=True`). Particles
                            that are in no FoF group have ID = NO_FOF_GROUP_ID =
                            np.array(-1,np.uintp).
        N_FoF (int):        The number of FoF groups found.
    '''
    if verbose is None:
        verbose = environment.verbose
    l = UnitScalar(l, s['pos'].units, subs=s, dtype=float)
    dvmax = UnitScalar(dvmax, s['vel'].units, subs=s, dtype=float)
    sort = bool(sort)
    min_N = int(min_N)

    if verbose >= environment.VERBOSE_NORMAL:
        print('perform a FoF search on %s particles:' % nice_big_num_str(len(s)))
        print('  l      = %.2g %s' % (l, l.units))
        if dvmax != np.inf:
            print('  dv_max = %.2g %s' % (dvmax, dvmax.units))
        print('  N     >= %g' % (min_N))
        sys.stdout.flush()

    pos = s['pos'].astype(np.float64)
    vel = s['vel'].astype(np.float64)
    mass = s['mass'].astype(np.float64)
    if pos.base is not None:
        pos = pos.copy()
    if vel.base is not None:
        vel = vel.copy()
    if mass.base is not None:
        mass = mass.copy()
    FoF = np.empty(len(s), dtype=np.uintp)

    # check setting of periodic boundaries
    periodic = False
    if periodic_boundary == 0:
        periodic = False
    elif periodic_boundary == 1:
        periodic = True
    elif periodic_boundary == 2:
        if s.cosmological:
            periodic = True
        else:
            periodic = False

    if periodic_boundary == 3:
        assert(boxsize_manual is not None)
        boxsize = float(boxsize_manual)
    else:
        if periodic:
            boxsize = float(s.boxsize.in_units_of(s['pos'].units))
        else:
            boxsize = float(s['pos'].in_units_of(s['pos'].units).max() * 2)

    C.cpygad.find_fof_groups(C.c_size_t(len(s)),
                             C.c_void_p(pos.ctypes.data),
                             C.c_void_p(vel.ctypes.data),
                             C.c_void_p(mass.ctypes.data),
                             C.c_double(l),
                             C.c_double(dvmax),
                             C.c_size_t(min_N),
                             C.c_int(int(sort)),
                             C.c_void_p(FoF.ctypes.data),
                             C.c_double(boxsize),
                             None,  # build new tree
                             )

    # do not count the particles with no halo!
    N_FoF = len(set(FoF)) - 1

    if verbose >= environment.VERBOSE_NORMAL:
        print('found %d groups' % N_FoF)
        N_list = min(N_FoF, 3)
        if N_list:
            if N_list == 1:
                print('the most massive one is:')
            else:
                print('the %d most massive ones are:' % N_list)
            for i in range(N_list):
                FoF_group = s[FoF == i]
                com = center_of_mass(FoF_group)
                M = FoF_group['mass'].sum()
                print('  group %d:   %8.3g %s  @  [%.3g, %.3g, %.3g] %s' % (
                    i, M, M.units, com[0], com[1], com[2], com.units))
        sys.stdout.flush()

    return FoF, N_FoF


_ROCKSTAR_HALO_DTYPES = [
    ('id', 'i'), ('internal_id', 'i'), ('num_p', 'i'),
    ('mvir', 'f'), ('mbound_vir', 'f'), ('rvir', 'f'), ('vmax', 'f'),
    ('rvmax', 'f'), ('vrms', 'f'), ('x', 'f'), ('y', 'f'), ('z', 'f'),
    ('vx', 'f'), ('vy', 'f'), ('vz', 'f'), ('Jx', 'f'), ('Jy', 'f'),
    ('Jz', 'f'), ('energy', 'f'), ('spin', 'f')
]
_ROCKSTAR_PART_DTYPES = [
    ('x', 'f'), ('y', 'f'), ('z', 'f'), ('vx', 'f'), ('vy', 'f'), ('vz', 'f'),
    ('particle_id', 'i'), ('assigned_internal_haloid', 'i'),
    ('internal_haloid', 'i'), ('external_haloid', 'i')
]


def Rockstar_halo_field_names():
    return [name for name, t in _ROCKSTAR_HALO_DTYPES]


def Rockstar_particle_field_names():
    return [name for name, t in _ROCKSTAR_PART_DTYPES]


class RockstarHeader(object):
    def __init__(self, txt):
        self._txt = txt

        props = {}
        for line in txt.split('\n'):
            for info in line.split(';'):
                info = info.split(':' if ':' in info else '=')
                if len(info) != 2 or not info[-1]:
                    # not a info with ':' or '=', or with an empty value
                    continue
                name, value = info
                name = name.strip('#')
                if name == 'Units':
                    try:
                        try:
                            name, value = value.split(' in ')
                        except:
                            name, value = value.split(' are ')
                    except:
                        continue
                props[name.strip()] = value.strip()
        self._props = props

    @property
    def txt(self):
        return self._txt

    @property
    def props(self):
        return self._props.copy()

    def get(self, key, default=None):
        return self._props.get(key, default)

    def __getitem__(self, key):
        return self.get(key)

    def __contains__(self, key):
        return key in self._props


def read_Rockstar_file(fname):
    '''
    Read in a Rockstar file.

    Note:
        Rockstar had to be run with the option 'FULL_PARTICLE_CHUNKS = 1', in
        order to output the whole particle data, which is needed here. This is
        also the file which filename needs to be specified.

    Args:
        fname (str):                The path to the particle(!) Rockstar output.

    Returns:
        header (RockstarHeader):    The header from the Rockstar file.
        halos (np.ndarray):         A numpy array with the halo table.
        particles (np.ndarray):     A numpy array with the particle table.
    '''
    from io import StringIO
    import codecs
    BEFORE_HALO_TBL = '#Halo table begins here:'
    BEFORE_PART_TBL = '#Particle table begins here:'

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('read Rockstar file "%s"' % fname)
    with codecs.open(fname, 'r', "utf-8") as f:
        rs_file = f.read()

    halo_start = rs_file.find(BEFORE_HALO_TBL)
    particle_start = rs_file.find(BEFORE_PART_TBL)

    # read the header
    header = RockstarHeader(rs_file[:halo_start])

    # read the halo table
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('  read halo table')
    # get rid of the leading '#'
    halo_tbl = '\n'.join(line[1:] \
                         for line in
                         rs_file[halo_start + len(BEFORE_HALO_TBL) + 1:particle_start].split('\n'))
    halos = np.loadtxt(StringIO(halo_tbl), dtype=_ROCKSTAR_HALO_DTYPES)

    # read the particle table
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('  read particle table')
    particles = np.loadtxt(
        StringIO(rs_file[particle_start + len(BEFORE_PART_TBL) + 1:]),
        dtype=_ROCKSTAR_PART_DTYPES)

    return header, halos, particles


def generate_Rockstar_halos(fname, snap, exclude=None, nosubs=True, data=None,
                            ignore_inconsistency=False, **kwargs):
    '''
    Create a list of halos from a Rockstar particle file.

    Args:
        fname (str):        The path to the particle(!) Rockstar output. (Cf. also
                            `read_Rockstar_file`!)
                            This argument is ignored, if `data` contains 'halos'
                            and 'particle'.
        snap (Snap):        The corresponding snapshot. Can be None, if used in
                            neither `exclude` nor calculation of the Halo classes
                            are requested.
        exclude (function): Exclude all halos h for which `exclude(h,snap)`
                            returns a true value. Here `h` is the halo information
                            from the Rockstar file, i.e. a np.ndarray with named
                            fields (for instance: `h['num_p']` is the number of
                            particles in that halos).
                            For all fields see `halo.Rockstar_halo_field_names()`.
        nosubs (bool):      TODO
        data (dict):        If a dictionary is passed, it will hold the Rockstar
                            tables for the halos and particles after the function
                            call.
                            Furthermore, if it already contains these two entries,
                            `fname` is ignored and this data is taken for the
                            generation.
        ignore_inconsistency (bool):
                            Just warn about inconsistencies between the Rockstar
                            file and the snapshot and do not raise exceptions.
        **kwargs            Further arguments are passed to `Halo.__init__`.

    Returns:
        halos (list):       A list of Halo class instances created from the
                            Rockstar output.
    '''
    # get the data / create dictionary to return them
    if data is not None and 'halos' in data and 'particles' in data:
        pass
    else:
        if data is None:
            data = {}
        # read the Rockstar file
        import time
        start_time = time.time()
        data['header'], data['halos'], data['particles'] = read_Rockstar_file(fname)
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('loaded in %.2f sec' % (time.time() - start_time))

    # check for consistency between Rockstar file and snapshot
    try:
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('checking for consistency...')
        h = data['header']
        if 'a' in h and abs(float(h['a']) - snap.scale_factor) > 0.01:
            raise RuntimeError("Scale factors of snapshot and Rockstar file do "
                               "not match: %s vs. %s" % (
                                   h['a'], snap.scale_factor))
        cosmo = snap.cosmology
        for name, attr in [('Om', 'Omega_m'), ('Ol', 'Omega_Lambda'), ('h', 'h_0')]:
            if name in h and abs(float(h[name]) - getattr(cosmo, attr)) > 0.0001:
                raise RuntimeError("%s of snapshot and Rockstar file " % (name, attr) +
                                   "do not match: %s vs. %s" % (
                                       h[name], getattr(cosmo, attr)))
        if 'Box size' in h:
            # it seems Rockstar supresses the information that the units are
            # comoving fot the box size
            bz = UnitScalar(str(h['Box size']).replace('h', 'h_0'),
                            snap.boxsize.units,
                            subs={'a': 1.0, 'z': 0.0, 'h_0': cosmo.h_0})
            if abs(float(bz / snap.boxsize) - 1) > 0.01:
                raise RuntimeError("box sizes of snapshot and Rockstar file " +
                                   "do not match: %s vs. %s" % (bz, snap.boxsize))
    except RuntimeError as e:
        if ignore_inconsistency:
            print('WARNING:', e.message)
        else:
            raise

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('create halo list from the Rockstar data')
    halo_classes = []
    parts = data['particles']

    halos = data['halos']
    import time
    start_time = time.time()
    # exclude non-physical halos and those excluded specifically
    halos = [h for h in data['halos']
             if ((h['id'] != -1) and not (exclude and exclude(h, snap)))]
    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('pre-selection from %s down to %s halos' % (
            len(data['halos']), len(halos)))

    # load blocks needed in order to avoid cluttering the progress bar
    if snap is not None:
        snap['ID']
        snap['pos']
        snap['mass']

    if 'calc' not in kwargs:
        kwargs['calc'] = None
    if 'properties' in kwargs:
        properties = kwargs['properties']
        del kwargs['properties']
    else:
        properties = {}
    start_time = time.time()
    with ProgressBar(halos, label='initialize halos', show_eta=False,
                     show_percent=False) as pbar:
        for h in pbar:
            # create mask for the particle data
            # 1st requirement: restrict to chosen halo
            # 2nd requirement: exclude substructures
            mask = (parts['external_haloid'] == h['id'])
            if nosubs:
                mask &= (parts['internal_haloid'] ==
                         parts['assigned_internal_haloid'])

            # enrich the halo properties (e.g. by units)
            prop = {dt[0]: val for dt, val in zip(_ROCKSTAR_HALO_DTYPES, h)}
            for name in ['x', 'y', 'z', 'vx', 'vy', 'vz', 'Jx', 'Jy', 'Jz']:
                del prop[name]
            prop.update({
                'mvir': UnitArr(h['mvir'], 'Msol/h_0'),
                'mbound_vir': UnitArr(h['mbound_vir'], 'Msol/h_0'),
                'center': UnitArr([h['x'], h['y'], h['z']], 'cMpc/h_0'),
                'vel': UnitArr([h['vx'], h['vy'], h['vz']], 'km/s'),
                'vmax': UnitArr(h['vmax'], 'km/s'),
                'rvmax': UnitArr(h['rvmax'], 'km/s'),
                'vrms': UnitArr(h['vrms'], 'km/s'),
                'J': UnitArr([h['Jx'], h['Jy'], h['Jz']],
                             'Msol/h_0 * Mpc/h_0 * km/s'),
                'energy': UnitArr(h['energy'], 'Msol/h_0 * (km/s)**2'),
            })
            # TODO: are these quantities those?
            # prop['mass'] = prop['mvir']
            # prop['com']  = prop['center']
            # (over-)write specified properties
            prop.update(properties)

            halo = Halo(IDs=parts['particle_id'][mask], root=snap,
                        properties=prop, **kwargs)
            halo_classes.append(halo)
    duration = time.time() - start_time

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('created a list of %s halos in %s' % (
            nice_big_num_str(len(halo_classes)), sec_to_nice_str(duration)))

    return halo_classes


class Halo(object):
    '''
    A class representing an halo or some group of an snapshot represented by IDs.

    Internally the defining IDs are stored and on creation some properties (such
    as total mass and R200) are calculated.

    Args:
        IDs (set, array-like):      The IDs that define the halo.
        halo (Snap):                A sub-snapshot that defines the halo. Only
                                    used if `IDs` is None.
        calc (set, list, tuple):    A list of the properties to calculate at
                                    instantiation.
                                    If calc='all', all defined properties are
                                    calculated.
                                    For an overview of the available properties
                                    see `Halo.calculable_props()` (also see
                                    `Halo.prop_descr(pro_name)`.
        root (Snap):                Some (sub-)snapshot that entirely contains the
                                    halo-defining IDs. It is only needed in case
                                    properties are about to calculate and `halo`
                                    is not given, otherwise it will be set to
                                    `halo.root`.
        properties (dict):          Properties to be set. This is done before the
                                    property calculation, meaning if a property
                                    would be calculated otherwise, it will not be
                                    if specified here.
        testing (bool):             Perform some (rather compute intensive)
                                    consistency checks.
    '''
    __long_name_prop__ = {
        'mass': 'total mass',
        'Mstars': 'total stellar mass',
        'Mgas': 'total gas mass',
        'Mdm': 'total dark matter mass',
        'parts': 'number of particles per species',
        'com': 'center of mass',
        'ssc': 'shrinking sphere center',
        'vel': 'mass-weighted velocity',
        'vel_sigma': 'mass-weighted velocity dispersion',
        'Rmax': 'maximum distance of a particle from com',
        'lowres_part': 'number of low-resolution particles',
        'lowres_mass': 'mass of low-resolution particles',
        'avg_gas_temp': 'mass weighted average gas temperature',
        'rho_gas_avg': 'mass weighted average gas density',
        'mean_stellar_age': 'mean stellar age',
        'gas_half_mass_radius_from_com': '3d half mass radius of gas from com',
        'gas_half_mass_radius_from_ssc': '3d half mass radius of gas from ssc',
        'stars_half_mass_radius_from_com': '3d half mass radius of stars from com',
        'stars_half_mass_radius_from_ssc': '3d half mass radius of stars from ssc'

    }

    for odens in ['vir', '200', '500']:
        prop = 'R' + odens
        __long_name_prop__[prop + '_FoF'] = \
            '%s (Mo+ 2002): (3*M / (4*pi * %s * rho_c))**(1/3.)' % (
                prop, '18*pi**2' if odens == 'vir' else odens)
        if odens == 'vir':
            continue
        for qty in ['R', 'M']:
            prop = qty + odens
            __long_name_prop__[prop + '_ssc'] = \
                'spherical %s with `virial_info` with ssc as center' % prop
            __long_name_prop__[prop + '_com'] = \
                'spherical %s with `virial_info` with com as center' % prop
        del prop, qty
    del odens

    @staticmethod
    def calculable_props():
        '''A list of the properties that can be calculated.'''
        return list(Halo.__long_name_prop__.keys())

    @staticmethod
    def prop_descr(name):
        '''Long description of a property.'''
        return Halo.__long_name_prop__.get(name, '<unknown>')

    def __init__(self, IDs=None, halo=None,
                 calc=None, root=None,
                 properties=None, testing=False):
        if IDs is None:
            if halo is None:
                raise ValueError('Halo needs to be defined by either an ID list '
                                 '(`IDs`) or a sub-snapshot (`halo`)!')
            IDs = halo['ID']

        # the defining property:
        self._IDs = np.array(IDs, copy=True)
        # property dictionary
        self._props = {}

        # add the predefined properties
        if properties is not None:
            for prop, val in properties.items():
                self._props[prop.replace(' ', '_')] = val

        # some testing for consistency
        if testing:
            IDset = set(self._IDs)
            if len(IDset) != len(self._IDs):
                raise ValueError('ID list is not unique!')
            if halo is None and root is not None:
                halo = root[IDMask(self._IDs)]
            if root is None and halo is not None:
                root = halo.root
            if root is not halo.root:
                raise ValueError('`halo` and `root` are inconsistent!')
            if len(IDset - set(root['ID'])) > 0:
                print('WARNING: Not all IDs are in the ' + \
                      'root snapshot!', file=sys.stderr)
        if halo is not None and len(halo) != len(self._IDs):
            print('WARNING: The given halo does not ' + \
                  'contain all specified IDs!', file=sys.stderr)

        # calculate the properties asked for

        if calc is None:
            return
        if halo is None:
            if root is None:
                raise ValueError('Need `halo` or `root` to calculate properties!')
            halo = root[IDMask(self._IDs)]
            if len(halo) != len(self._IDs):
                print('WARNING: The given snapshot does not ' + \
                      'contain all specified IDs!', file=sys.stderr)
        elif root is not None and testing:
            if root is not halo.root:
                print('WARNING: `halo` and `root` are ' + \
                      'inconsistent!', file=sys.stderr)
        if root is None:
            root = halo.root

        if calc == 'all':
            calc = Halo.calculable_props()
        for prop in calc:
            self.calc_prop(prop, halo=halo, root=root, recompute=False)

    @property
    def IDs(self):
        '''The defining IDs.'''
        return self._IDs

    @property
    def mask(self):
        '''The ID mask of the halo to mask a snapshot.'''
        # avoid copying the IDs
        mask = IDMask.__new__(IDMask)
        SnapMask.__init__(mask)
        mask._IDs = self._IDs
        return mask

    @property
    def props(self):
        '''All properties.'''
        return self._props.copy()

    def __dir__(self):
        return list(self.__dict__.keys()) + list(self._props.keys()) + \
               [k for k in list(self.__class__.__dict__.keys())
                if not k.startswith('_')]

    def __getattr__(self, name):
        if name.startswith('__'):
            return super(Halo, self).__getattr__(name)
        attr = self._props.get(name, None)
        if attr is not None:
            return attr
        raise AttributeError('%s has no attribute "%s"!' % (self, name))

    def __repr__(self):
        r = '<Halo @%s' % hex(id(self))
        try:
            r += ', N = %s' % nice_big_num_str(len(self))
        except:
            pass
        try:
            r += ', M = %.2g %s' % (self.mass, self.mass.units)
        except:
            pass
        r += '>'
        return r

    def __len__(self):
        if 'parts' in self._props:
            return sum(self._props['parts'])
        else:
            return None

    def _get(self, prop, halo=None, root=None, recompute=False):
        if recompute or prop not in self._props:
            self.calc_prop(prop, halo=halo, root=root)
        return self._props[prop]

    def calc_prop(self, prop='all', halo=None, root=None, recompute=True):
        '''
        Calculate a (defined) property.

        Args:
            prop (str):         The property's name to compute. If it is 'all',
                                all properties in `Halo.calculable_props()` are
                                computed.
            halo (Snap):        The (sub-)snapshot of this halo, i.e. `root[self]`
                                where `root` is the root snapshot.
            root (Snap):        The root snapshot of this halo. Does not need to
                                be specified, but if `halo` is None, it will be
                                set by masking this root with the IDMask of this
                                halo.
            recompute (bool):   Recompute this quantity even if it already is
                                stored in the properties dictionary.

        Returns:
            val (...):      The value of the property calculated (or just taken
                            from the properties dictionary, if recompute=False).
                            In case of calc='all', the entire property dictionary
                            is returned.
        '''
        # preparation
        if not recompute and (isinstance(prop, str) and prop != 'all'
                              and prop in self._props):
            return self._props[prop]
        if halo is None:
            halo = root[self.mask]
        elif root is None:
            root = halo.root
        args = {'halo': halo, 'root': root, 'recompute': recompute}

        # handle special cases
        if prop == 'all':
            for p in Halo.calculable_props():
                self.calc_prop(p, **args)
            return self.props
        if isinstance(prop, (list, tuple, np.ndarray)):
            for p in prop:
                self.calc_prop(p, **args)
            return self.props

        # compute the requested
        if prop == 'mass':
            val = halo['mass'].sum()
        elif prop == 'com':
            val = center_of_mass(halo)
        elif prop == 'parts':
            val = tuple(halo.parts)  # shall not change!
        elif prop in ['Mstars', 'Mgas', 'Mdm']:
            sub = getattr(halo, prop[1:], None)
            if sub is None:
                val = UnitScalar(0.0, halo['mass'].units)
            else:
                val = sub['mass'].sum()
        elif prop == 'vel':
            val = mass_weighted_mean(halo, 'vel')
        elif prop == 'vel_sigma':
            v0 = mass_weighted_mean(halo, 'vel')
            val = np.sqrt(np.sum(mass_weighted_mean(halo, 'vel**2') - v0 ** 2))
        elif prop == 'ssc':
            com = self._get('com', **args)
            R_max = np.percentile(periodic_distance_to(halo['pos'],
                                                       com,
                                                       halo.boxsize),
                                  90)
            val = shrinking_sphere(halo, center=com,
                                   R=R_max, verbose=False)
        elif prop == 'Rmax':
            val = periodic_distance_to(halo['pos'], self._get('com', **args),
                                       halo.boxsize).max()
        elif prop[0] in ['R', 'M'] and len(prop) > 4 and \
                (prop[1:4] == 'vir' or prop[1:4].isdigit()) and prop[4] == '_':
            qty = prop[0]
            odens_n = prop[1:4]
            scheme = prop[5:]
            odens = 18. * np.pi ** 2 if odens_n == 'vir' else float(odens_n)
            if scheme == 'FoF':
                rho_crit = halo.cosmology.rho_crit(z=halo.redshift)
                rho_crit.convert_to(halo['mass'].units / halo['pos'].units ** 3,
                                    subs=halo)
                mass = self._get('mass', **args)
                r3 = 3.0 * mass / (4.0 * np.pi * odens * rho_crit)
                val = r3 ** Fraction(1, 3)
            else:
                if scheme not in ['com', 'ssc']:
                    raise ValueError('Unknown scheme for property "%s"!' % prop)
                center = self._get(scheme, **args)
                Rodens, Modens = virial_info(root, center=center, odens=odens)
                val = Rodens if qty == 'R' else Modens
                # don't waste the additional information!
                for qty in ['R', 'M']:
                    name = qty + odens_n + '_' + scheme
                    # if we recompute some property, it might be requested to keep
                    # old values...
                    if not recompute and name not in self._props:
                        self._props[name] = Rodens if qty == 'R' else Modens
        elif prop == 'lowres_part':
            low = getattr(halo, 'lowres', None)
            val = 0 if low is None else len(low)
        elif prop == 'lowres_mass':
            low = getattr(halo, 'lowres', None)
            if low is None:
                val = UnitScalar(0.0, halo['mass'].units)
            else:
                val = low['mass'].sum()
        elif prop == 'avg_gas_temp':
            val = mass_weighted_mean(halo.gas, 'temp')
        elif prop == 'rho_gas_avg':
            val = mass_weighted_mean(halo.gas, 'rho')
        elif prop == 'mean_stellar_age':
            # only if there are stars in the halo
            if halo.stars['mass'].size > 0:
                val = halo.stars['age'].mean()
            else:
                val = np.nan
        elif prop == 'gas_half_mass_radius_from_com':
            val = half_mass_radius(halo.gas, M=halo.gas['mass'].sum(), center=self._get('com', **args), proj=None)
        elif prop == 'gas_half_mass_radius_from_ssc':
            val = half_mass_radius(halo.gas, M=halo.gas['mass'].sum(), center=self._get('ssc', **args), proj=None)
        elif prop == 'stars_half_mass_radius_from_com':
            if halo.stars['mass'].size > 0:
                val = half_mass_radius(halo.stars, M=halo.stars['mass'].sum(), center=self._get('com', **args), proj=None)
            else:
                val = np.nan

        elif prop == 'stars_half_mass_radius_from_ssc':
            if halo.stars['mass'].size > 0:
                val = half_mass_radius(halo.stars, M=halo.stars['mass'].sum(), center=self._get('ssc', **args), proj=None)
            else:
                val = np.nan

        else:
            raise ValueError('Unknown property "%s"!' % prop)

        self._props[prop] = val
        return val


def nxt_ngb_dist_perc(s, q, N=1000, tree=None, ret_sample=False, verbose=None):
    '''
    Estime the percentile distance to the the next neighbour within the given snapshot.

    Args:
        s (Snap):           The snapshot to use.
        q (int,float):      The percentile to ask for.
        N (int):            The sample size to estimate the distance from.
        tree (cOctree):     The octree class to use, if already present. Will be
                            generated on the fly otherwise.
        ret_sample (bool):  Also return the entire sample drawn.
        verbose (int):      Verbosity level. Default: the gobal pygad verbosity
                            level.
    Returns:
        d (UnitArr):        The q-percentile distance to the next neighbour.
       [dists (UnitArr):    The sample drawn.]
    '''
    if verbose is None:
        verbose = environment.verbose
    from .. import octree
    pos = s['pos'].view(np.ndarray)
    boxsize = s.boxsize.in_units_of(s['pos'].units)
    if tree is None:
        if verbose >= environment.VERBOSE_NORMAL:
            print('building the octree...')
        tree = octree.cOctree(pos)
    if verbose >= environment.VERBOSE_NORMAL:
        print('preparing...')
    d = np.empty(N, dtype=float)
    cond = np.ones(len(s), dtype=np.int32)
    with ProgressBar(np.random.randint(len(s), size=N), label='sampling') as pbar:
        for i in pbar:
            cond[i] = 0
            i_next = tree.find_next_ngb(pos[i], pos, periodic=boxsize, cond=cond)
            cond[i] = 1
            d[pbar.iteration - 1] = dist(pos[i], pos[i_next])
    if ret_sample:
        return UnitArr(np.percentile(d, q), s['pos'].units), \
               UnitArr(d, s['pos'].units)
    else:
        return UnitArr(np.percentile(d, q), s['pos'].units)


def generate_FoF_catalogue(s, l=None, calc='all', FoF=None, exclude=None,
                           max_halos=None, ret_FoFs=False, verbose=None,
                           progressbar=True, **kwargs):
    '''
    Generate a list of Halos defined by FoF groups.

    Args:
        s (Snap):           The snapshot to generate the FoF groups for.
        l (UnitScalar):     The linking length for the FoF groups (for dark
                            matter / all matter it turned out that
                            ( 1500. * rho_crit / median(mass) )^(-1/3) is a
                            reasonable value, which will be used as default, when
                            l=None).
        calc (set, list, tuple):
                            A list of the properties to calculate at instantiation
                            of the Halo instances. (Cf. also `Halo`.)
        FoF (np.array):     An array with group IDs for each particle in the
                            snapshot `s`. If given, `l` is ignored and the FoF
                            groups are not calculated within this function.
        exclude (function): Exclude all halos h for which `exclude(h,s)` returns
                            a true value (note: s[h] gives the halo as a
                            sub-snapshot).
        max_halos (int):    Limit the number of returned halos to the `max_halos`
                            most massive halos. If None, all halos are returned.
        ret_FoFs (bool):    Also return the array with the FoF group indices
                            (sorted by mass).
        verbose (bool):     Verbosity level. Default: the gobal pygad verbosity
                            level.
        progressbar (bool): Whether to show a progress bar for initialising the
                            halo classes (not for generating the FoF groups,
                            though!).
        **kwargs:           Other keywords are passed to `find_FoF_groups` (e.g.
                            `dvmax`).

    Returns:
        halos (list):       A list of all the groups as Halo instances. (Sorted in
                            mass, it not `sort=False` in the `kwargs`.)

        if ret_FoFs==True:
        FoF (np.array):     The FoF group IDs for each particle; particles with no
                            group have ID = NO_FOF_GROUP_ID (cf. `FoF` argument
                            and/or function `find_FoF_groups`).
        N_FoF (int):        The number of FoF groups in `FoF`.
    '''
    if verbose is None:
        verbose = environment.verbose
    if FoF is None:
        if l is None:
            l = (1500. * s.cosmology.rho_crit(s.redshift)
                 / np.median(s['mass'])) ** Fraction(-1, 3)
        FoF, N_FoF = find_FoF_groups(s, l=l, verbose=verbose, **kwargs)
    else:
        N_FoF = len(set(FoF)) - 1

    from ..utils import ProgressBar, DevNull
    if verbose >= environment.VERBOSE_NORMAL and progressbar:
        outfile = sys.stdout
    else:
        outfile = DevNull()
    halos = []
    with ProgressBar(
            range(min(N_FoF, max_halos) if exclude is None else N_FoF),
            show_eta=False,
            show_percent=False,
            label='initialize halos',
            file=outfile) as pbar:
        if not progressbar:
            print('initialize halos from FoF group IDs...')
            sys.stdout.flush()
        for i in pbar:
            h = Halo(halo=s[FoF == i], root=s, calc=calc)
            h.linking_length = l
            if exclude is None or not exclude(h, s):
                halos.append(h)
            if len(halos) == max_halos:
                break

    if verbose >= environment.VERBOSE_NORMAL:
        print('initialized %d halos.' % (len(halos)))
        sys.stdout.flush()

    if ret_FoFs:
        return halos, FoF, N_FoF
    else:
        del FoF
        return halos


def _common_mass(h1, h2, s):
    # access private attribute for speed
    ID_both = set(h1._IDs) & set(h2._IDs)
    return s[IDMask(ID_both)]['mass'].sum()


def find_most_massive_progenitor(s, halos, h0):
    '''
    Find the halo with the most mass in common.

    Args:
        s (Snap):       The snapshot to which the halo catalogue `halos` belongs.
        halos (list):   A list of halos in which the one with the most common mass
                        with `h0` is searched for.
        h0 (Halo):      The halo for which to find the most massive progenitor in
                        `halos` (the one with the most mass in common).

    Returns:
        mmp (Halo):     The most massive progenitor (in `halos`) of `h0`.
    '''
    if len(halos) == 0:
        return None

    h0_mass = h0.mass.in_units_of(halos[0].mass.units, subs=s)

    # propably one of the closest halos (with at least 10% of the mass), find them
    closest = []
    min_mass = 0.1 * h0_mass
    min_mass.convert_to(halos[0].mass.units, subs=s)
    h0_com = h0.com.in_units_of(halos[0].com.units, subs=s)
    for i in range(3):
        close = None
        close_d = np.inf
        for h in halos:
            d = np.linalg.norm(h.com - h0_com)
            if d < close_d and h.mass > min_mass:
                for nh in closest:
                    if h is nh:
                        continue
                close = h
                close_d = d
        if close is not None:
            closest.append(close)

    # if any of them has more than 50% of the mass, we are done
    com_mass = []
    for h in closest:
        com_mass.append(_common_mass(h, h0, s))
        if com_mass[-1] / h0_mass > 0.5:
            return h

    # if no other halo can have more mass than the most massive of the closest, it
    # is this one
    if closest:
        mm_closest, cm = max(list(zip(closest, com_mass)), key=lambda p: p[1])
        if h0_mass - sum(com_mass) < cm:
            return mm_closest

    # iterate and find the most massive progenitor
    # not done in the beginning, since this requires the calculation of the common
    # mass for *all* halos and, hence, is slow
    mmp = max(halos,
              key=lambda h: _common_mass(h, h0, s))
    return mmp if _common_mass(mmp, h0, s) > 0 else None




# these packages are loaded by gcache3 by default
import os
import io
import sys
import re
import time
import numpy as np
import matplotlib.pyplot as plt
import pickle as pickle
import pygad as pg

#############################################################################
# additional packages should be loaded in prepare section                   #
# here your code of functions ONLY starts                                   #
#############################################################################

def prepare_snap(snap_cache, cmd_par, cmd_par1):
    # snapshot is already translated and orientated by SnapshotCache
    # ##############################################################
    # snap_num_freq = args.snaps.count('%')
    # snapfile = args.snaps % ((snap_num,) * snap_num_freq)
    # snap = pygad.Snap(snapfile,
    #                   physical=False,
    #                   gad_units={
    #                     'LENGTH':   args.length,
    #                     'VELOCITY': args.velocity,
    #                     'MASS':     args.mass,
    #                   })
    # snap_num_freq = args.infos.count('%')
    # infofile = args.infos % ((snap_num,) * snap_num_freq)
    # info = pygad.tools.read_info_file(infofile)
    #
    # if args.verbose:
    #     print('  ->', snap)
    #     print('  scale param.:', snap.scale_factor)
    #     print('  cosmic time: ', snap.cosmic_time())
    #
    # Ired = info['I_red(gal)']
    # if Ired is None or info['center'] is None:
    #     if args.verbose: print('no central galaxy yet; continue...')
    #     return snap, None
    # Ired = Ired.reshape((3,3))
    #
    # if args.verbose:
    #     print('center (also in velocity) and orientate the snapshot...')
    # if args.verbose:
    #     print('   center at:', info['center'])
    # pygad.Translation(-info['center']).apply(snap)
    # rind = np.argsort(snap['r'])
    # snap['vel'] -= pygad.analysis.mass_weighted_mean(
    #                 snap[snap['r'] < snap['r'][rind[1001]]], 'vel')
    # if args.verbose:
    #     print('   orientation is at reduced inertia tensor')
    # pygad.analysis.orientate_at(snap, 'red I', qty=Ired, total=True)

    snap = snap_cache.snapshot
    snap.to_physical_units()       # TODO not yet clear

    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('restrict to region...')
    if cmd_par == 'disc':
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("  ... the disc (jz/jc>%.2f," % cmd_par1, end=' ')
            print("rmax=0.15*R200, zmax='5 kpc')")
            print("  ... where jc is the angular momentum of a particle with", end=' ')
            print("the same energy, but on a perfectly circular orbit; only", end=' ')
            print("co-rotating particles are counted")
        R200, M200 = pg.analysis.virial_info(snap)
        region = snap[ pg.DiscMask(jzjc_min=cmd_par1,
                                      rmax=0.15*R200, zmax='5 kpc') ]
    elif cmd_par == 'ball':
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("  ... all within %g%% of R200" % (100.*cmd_par1))
        R200, M200 = pg.analysis.virial_info(snap)
        region = snap[ pg.BallMask(cmd_par1*R200) ]
    elif cmd_par == 'disc-Uebler':
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print("  ... the disc (|jz/jc| in [0.8,1.2], rmax=0.15*R200, zmax='3 kpc')")
            print("  ... where jc = r * sqrt(G*M(<r)/r)")
        R200, M200 = pg.analysis.virial_info(snap)
        r_sort_arg = np.argsort(snap['r'])
        M_enc = np.cumsum( snap['mass'][r_sort_arg] )\
                [pg.utils.perm_inv(r_sort_arg)]
        jc = snap['mass'] * snap['r'] * np.sqrt( pg.physics.G * M_enc / snap['r'] )
        del r_sort_arg, M_enc
        jz_jc = (snap['angmom'][:,2]/jc).in_units_of(1,subs=snap)
        rmax = (0.15*R200).in_units_of(snap['r'].units, subs=snap)
        zmax = pg.UnitArr('3 kpc').in_units_of(snap['z'].units, subs=snap)
        region = snap[ (0.8 < np.abs(jz_jc)) & (np.abs(jz_jc) < 1.2) &
                       (snap['r'] < rmax) & (snap['z'] < zmax) ]
        del jc, jz_jc
    else:
        raise RuntimeError('Unkown region mode "%s"!' % cmd_par)

    return snap, region


def add_state_info( sub, new=False, vel=False, temp=True, info='' ):
    dtypel = [('ID', 'uint64'), ('a', 'float64'), ('mass', 'float64'),
              ('metals', 'float64'), ('angmom','float64')]
    if temp:
        dtypel.append(('temp', 'float64'))
    if vel:
        dtypel.append(('vel','float64'))

    data = np.zeros((len(sub),), dtype=np.dtype(dtypel))
    data['ID'] = sub['ID']
    data['a'] = sub.scale_factor * np.ones(len(sub),float)
    data['mass'] = sub['mass']
    data['metals'] = sub['metals']
    data['angmom'] = sub['angmom'][:,2]
    # data = [sub['ID'],
    #         sub.scale_factor * np.ones(len(sub),float),
    #         sub['mass'],
    #         sub['metals'],
    #         sub['angmom'][:,2]]
    if temp:
        # data.append( sub['temp'] )
        if 'temp' in sub.available_blocks():
            data['temp'] = sub['temp']
    if vel:
        # data.append( np.linalg.norm(sub['vel'],axis=-1) )
        if 'vel' in sub.available_blocks():
            data['vel'] = np.linalg.norm(sub['vel'],axis=-1)

    # data = np.vstack(data)
    # data = np.transpose(data)
    if new:
        for row in data:
            trace_data[int(row[0])] = [('n', list(row)[1:])]
    else:
        for row in data:
            ID = int(row[0])
            #trace_data[ID] = np.concatenate( (trace_data[ID], list(row)[1:]) )
            trace_data[ID].append((info, list(row)[1:]))
            #print("append add_state_info ",ID, len(list(row)[1:]), " ", info)
    return


def add_state_missing( missing_IDs, sub):
    dtypel = [('ID', 'uint64'), ('a', 'float64')]

    data = np.zeros((len(missing_IDs),), dtype=np.dtype(dtypel))
    data['ID'] = np.array(list(missing_IDs), dtype=np.uint64)
    data['a'] = sub.scale_factor * np.ones(len(missing_IDs),float)
    for row in data:
        ID = int(row[0])
        trace_data[ID].append(('v', list(row)[1:]))
    return


def start_trace_new( region, region_IDs ):
    '''Start to trace particles that are new in region.'''
    new_IDs = region_IDs - traced_IDs
    traced_IDs.update(new_IDs)
    new = region.gas[ pg.IDMask(new_IDs) ]
    if len(set(new['ID'])) - len(set(new_IDs)) != 0:
        print("FEHLER start_trace_new Mask ", len(set(new['ID'])), len(set(new_IDs)) )
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('new gas particles to trace:             %-10d' % len(new))
    add_state_info( new, new=True, info="n" )
    return


def stop_trace_stars( stars ):
    '''Stop tracing particles that turned into stars.'''
    new_star_IDs = set(stars['ID']).intersection(traced_IDs)
    traced_IDs.difference_update( new_star_IDs )
    new_stars = snap.stars[ pg.IDMask(new_star_IDs) ]
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('traced gas particles that formed stars: %-10d' % len(new_stars))
    add_state_info( new_stars, temp=False, info="s"  )
    return


def deactivate_reentering( region, region_IDs ):
    reenter_IDs = out_IDs.intersection(region_IDs)
    out_IDs.difference_update( reenter_IDs )
    reenter = region[ pg.IDMask(reenter_IDs) ]
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('gas particles re-entered the region:    %-10d' % len(reenter))
    add_state_info( reenter , info="r")
    return


def activate_leaving( snap, region_IDs ):
    leaving_IDs = (traced_IDs - out_IDs) - region_IDs
    out_IDs.update( leaving_IDs )
    leaving = snap.gas[ pg.IDMask(leaving_IDs) ]
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('gas particles leaving the region:       %-10d' % len(leaving))
    add_state_info( leaving, vel=True, info="l" )
    missing_IDs = leaving_IDs - set(leaving['ID'])
    add_state_missing(missing_IDs, leaving)

    dtypel = [('ID', 'uint64'), ('a1', 'float64'), ('r', 'float64'),
              ('a2', 'float64'), ('pos','float64')]
    data = np.zeros((len(leaving),), dtype=np.dtype(dtypel))
    data['ID'] = leaving['ID']
    data['a1'] = leaving.scale_factor * np.ones(len(leaving),float)
    data['r'] = leaving['r']
    data['a2'] = leaving.scale_factor * np.ones(len(leaving),float)
    data['pos'] = leaving['pos'][:,2]
    # data = [leaving['ID'],
    #         leaving.scale_factor * np.ones(len(leaving),float),
    #         leaving['r'],
    #         leaving.scale_factor * np.ones(len(leaving),float),
    #         leaving['pos'][:,2]]
    # data = np.vstack( data ).T
    for row in data:
        ID = int(row[0])
        #trace_data[ID] = np.concatenate( [trace_data[ID], list(row)[1:]] )
        trace_data[ID].append(('p', list(row)[1:]))
        #print("Append activate_leaving ", ID, len(list(row)[1:]))
    return


def update_activated( snap ):
    active = snap.gas[ pg.IDMask(out_IDs) ]
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('update traced particles outside:        %-10d' % len(out_IDs))

    dtypel = [('ID', 'uint64'), ('a1', 'float64'), ('r', 'float64'),
              ('a2', 'float64'), ('pos', 'float64')]
    current = np.zeros((len(active),), dtype=np.dtype(dtypel))
    current['ID'] = active['ID']
    current['a1'] = active.scale_factor * np.ones(len(active),float)
    current['r'] = active['r']
    current['a2'] = active.scale_factor * np.ones(len(active),float)
    current['pos'] = active['pos'][:,2]
    # current = [active['ID'],
    #            active.scale_factor * np.ones(len(active),float),
    #            active['r'],
    #            active.scale_factor * np.ones(len(active),float),
    #            np.abs(active['pos'][:,2])]
    # current = np.vstack( current ).T
    for row in current:
        data = trace_data[int(row[0])]
        if data[-3] < list(row)[-3]:
            data[-4:-2] = list(row)[-4:-2]
        if data[-1] < list(row)[-1]:
            data[-2:] = list(row)[-2:]
        #print("Concatenate update_activated ", row[0], len(data))
    return


def get_trace_file_name():
    return snap_cache.get_profile_path() + cmd_par2

##############################################################################
# here start the gcache3 sections, invoking your code                        #
##############################################################################
# the order of the gcache3 sections is : init, prepare, process, close, end  #
# sections are identified by line starting with # gcache3-<section name>:    #
# the order must not be modified                                             #
##############################################################################

##############################################################################
# gcache3-init: ignored by gcache3
##############################################################################
pg.environment.verbose=pg.environment.VERBOSE_QUIET
os.environ['SNAPSHOT_HOME'] = '/mnt/hgfs/AstroDaten/CosmoZooms/'
os.environ['SNAPCACHE_HOME'] = '/mnt/hgfs/Astro/CosmoCache/'

snap_fname_rel = "M0094/AGN/snap_m0094_sf_x_2x_060"
snap_cache = pg.SnapshotCache(snap_fname_rel, profile='FullRun')
snap_cache.load_snapshot()

snap_num = 94
snap_start = 94
snap_end = 94
snap_overwrite = True

# parser.add_argument('--region',
#                     choices=['disc', 'ball', 'disc-Uebler'],
#                     default='ball',
#                     help='The definition of the region (default: %(default)s). ' +
#                          '"disc" is everything with jz/jc>0.85, r<0.15*R200, ' +
#                          'and |z|<5 kpc, where jc is the angular momentum of a ' +
#                          'particle on a circular orbit with the same energy, ' +
#                          'only co-rotating particles are counted; "ball" is ' +
#                          'just everything within 0.10*R200; "disc-Uebler" is ' +
#                          'the definition of Uebler+ (2014): 0.8<|jz/j\'c|<1.2, ' +
#                          'r<0.15*R200, and |z|<5 kpc, where ' +
#                          'j\'c := r * sqrt(G*M(<r)/r).')
# parser.add_argument('--region_param',
#                     default=None,
#                     help='A parameter for the region. For "disc" it is the ' +
#                          'critical jz/jc value (default: 0.85); for "ball" it ' +
#                          'is the radius in terms of the virial radius ' +
#                          '(default: 10%%), for "disc-Uebler" it is ignored.')
cmd_par = 'gx'      # region
cmd_par1 = '0.1'     # region_param
cmd_par2 = 'traced_gas.dat'

##############################################################################
# gcache3-prepare: processes before first snapshot
##############################################################################

# Parameter not yet clear, how to migrate
# parser.add_argument('--length',
#                     default='ckpc/h_0',
#                     help='Define the Gadget length units (default: %(default)s).')
# parser.add_argument('--mass',
#                     default='1e10 Msol h_0**-1',
#                     help='Define the Gadget mass units (default: %(default)s).')
# parser.add_argument('--velocity',
#                     default='km/s',
#                     help='Define the Gadget velocity units (default: %(default)s).')

print("command CmdTraceGas - prepare ", snap_num, " von ", snap_start, " bis ", snap_end)
#pg.environment.verbose = pg.environment.VERBOSE_TACITURN
if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
    print('starting up...')

#trace_file_name = get_trace_file_name()
#if os.path.exists(trace_file_name) and not snap_overwrite:
#    print('ERROR: destination path exsits ' + \
#       '(%s)! Consider "--overwrite".' % trace_file_name, file=sys.stderr)
#    sys.exit(1)

# trace (gas) particles by ID
trace_data = {}
traced_IDs = set()
out_IDs = set()
snap = None

globals()['trace_data'] = trace_data
globals()['traced_IDs'] = traced_IDs
globals()['out_IDs'] = out_IDs
globals()['snap'] = snap

if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
    start_total = time.time()

# processing some arguments
if cmd_par is None:
    cmd_par = 'gx'
if cmd_par == 'disc':
    if cmd_par1 is None: cmd_par1 = 0.85
    cmd_par1 = float(cmd_par1)
elif cmd_par == 'ball':
    if cmd_par1 is None: cmd_par1 = 0.1
    cmd_par1 = float(cmd_par1)
elif cmd_par == 'disc-Uebler':
    pass  # no use of cmd_par1
elif cmd_par == 'gx':
    pass # region of galaxy as set by SnapshotCache
else:
    raise RuntimeError('Unkown region mode "%s"!' % cmd_par)

# Suspend importing to here, after parsing the arguments to be quick,
# if only the help has to be displayed or there already occurs and
# error in parsing the arguments.

if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
    print('imported pygad', pg.version)

# not necessary, covered by gCache
# if os.path.exists(args.dest) and not args.overwrite:
#     print('ERROR: destination path exsits ' + \
#           '(%s)! Consider "--overwrite".' % args.dest, file=sys.stderr)
#     sys.exit(1)

##############################################################################
# gcache3-process: processes for each snapshot
##############################################################################
print("command CmdTraceGas - process", snap_num)

start_loop = time.time()
if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
    print('====================================================')
    print('read snapshot and trace file #%03d...' % snap_num)
    sys.stdout.flush()
if cmd_par == 'gx':
    snAllIDs = snap_cache.snapshot['ID']
    snIDs, snCnt = np.unique(snAllIDs,return_counts=True)
    snDict = dict([ (ID, Cnt) for ID, Cnt in zip(snIDs, snCnt)])
    snUMask =np.array([(snDict[ID] == 1) for ID in snAllIDs], dtype=bool)
    snap = snap_cache.snapshot[snUMask]
    if snap_cache.galaxy is None:
        region = None
    else:
        gxAllIDs = snap_cache.galaxy['ID']
        gxIDs, gxCnt = np.unique(gxAllIDs,return_counts=True)
        gxDict = dict([ (ID, Cnt) for ID, Cnt in zip(gxIDs, gxCnt)])
        gxUMask = np.array([(gxDict[ID] == 1 and snDict[ID] == 1) for ID in gxAllIDs], dtype=bool)
        region = snap_cache.galaxy[gxUMask]
else:
    snap, region = prepare_snap(snap_cache, cmd_par, cmd_par1)

if region is not None:  # central galaxy
    region_IDs = set(region.gas['ID'])

    if len(set(region_IDs) - set(snap['ID'])) > 0:
        print("FEHLER Region_IDs - ", set(region_IDs) - set(snap['ID']))

    # continue as long as there is not / has not been a region with
    # at least 100 particles
    if not trace_data and len(region.baryons) < 100:
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('no region yet (only considered defined when it has more')
            print('than 100 baryonic particles)')
            print('continue')
    else:
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print(r'region has a stellar mass of %.2g %s (%d particles)' % (
                region.stars['mass'].sum(), region.stars['mass'].units,
                len(region.stars)))
            print(r'   ... and a gas mass of %.2g %s (%d particles)' % (
                region.gas['mass'].sum(), region.gas['mass'].units, len(region.gas)))
            sys.stdout.flush()
        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('currently tracing %d particles...' % len(traced_IDs))

        start_trace_new(region, region_IDs)
        stop_trace_stars(snap.stars)

        if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
            print('... now tracing %d particles' % len(traced_IDs))

        deactivate_reentering(region, region_IDs)
        activate_leaving(snap, region_IDs)
        #update_activated(snap)

if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
    print('done in %.3gs' % (time.time() - start_loop))

snap_cache.gx_properties.append('gashistoryfile', cmd_par2)

print("command CmdTraceGas - process ready", snap_num, " len=", len(trace_data))

##############################################################################
# gcache3-close: processed after last snapshot
##############################################################################
print("command CmdTraceGas - close")
if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
    print('====================================================')
    print('store results...')
    sys.stdout.flush()

gas_trace_filename = get_trace_file_name()
with open(gas_trace_filename, 'wb') as f:
    pickle.dump(trace_data, f, pickle.HIGHEST_PROTOCOL)


##############################################################################
# gcache3-end: ignored by gcache3
##############################################################################





'''
A collection of functions that calculate properties of snapshots (Snap) and
sub-snapshots (e.g. MaskedSnap).
'''
__all__ = ['radial_surface_density_profile']

from units import UnitArr
import physics
import numpy as np
import sph

def radial_surface_density_profile(snap, quantity='mass', r_edges=None,
                                   N_sample=50, units=None, verbose=True):
    '''
    Create a radial surface density profile.

    Args:
        snap (Snap):        The snapshot to use.
        quantity (str, UnitArr):
                            The quantity to do the radial surface density profile
                            of.
        r_edges (array-like):
                            The edges of the radial bins. If no units are set,
                            they are assumed to be those of snap.pos. (default:
                            UnitArr(np.arange(51), units='kpc'))
        units (str, Unit):  The units of the surface density.
        verbose (bool):     Whether or not to print which sample is mapped at the
                            moment.

    Returns:
        Sigma (UnitArr):   The profile, i.e. the quantity summed for the
                            intervals defined by r_edges and then divided by the
                            area.
    '''
    if r_edges is None:
        r_edges = UnitArr(np.arange(51), units='kpc')
    elif not isinstance(r_edges, UnitArr):
        r_edges = UnitArr(r_edges)
    if r_edges._units is None:
        r_edges.units = snap.pos.units

    Sigma = profile(snap, quantity, r_edges=r_edges, mode='cylindrical',
                    average='none', N_sample=N_sample, verbose=verbose)
    for i in range(len(r_edges)-1):
        Sigma[i] /= np.pi*(r_edges[i+1]**2-r_edges[i]**2)
    Sigma.units /= r_edges.units**2
    if units is not None:
        Sigma.convert_to(units)

    return Sigma



'''
This module reads and interpolates Bruzual & Charlot single stellar population
(SSP) model tables.

Examples:
    >>> ages, u_b = load_table('U-B', 'm42')
    >>> u_b[::10]
    array([-1.0793, -1.0793, -1.0793, -1.0793, -1.0082, -0.6092, -0.6533,
           -0.7324, -0.7617, -0.5991, -0.5185, -0.3466, -0.1312,  0.0532,
            0.1679,  0.227 ,  0.2277,  0.2713,  0.2879,  0.2926,  0.2845,
            0.275 ])
    >>> for Z in [0.001, 0.01, 0.03]:
    ...     print(inter_bc_qty('1.23 Myr', Z, qty='Mbol'),
    ...           inter_bc_qty('15.2 Myr', Z, qty='Mbol'),
    ...           inter_bc_qty('0.8 Gyr', Z, qty='Mbol'),
    ...           inter_bc_qty('2.5 Gyr', Z, qty='Mbol'))
    -2.617717e+00 -5.640066e-01 3.811986460678126 4.700866682636425
    -2.657609e+00 -4.655381e-01 3.9231958298509078 4.904316681049632
    -2.784935e+00 -2.826923e-01 4.045632843503789 5.05050001420477

    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_320', physical=True)
    >>> inter_bc_qty(s.stars['age'], s.stars['metallicity'], qty='log Nly')
    load block form_time... done.
    derive block age... done.
    load block Z... done.
    derive block elements... done.
    derive block H... done.
    derive block He... done.
    derive block metals... done.
    derive block metallicity... done.
    UnitArr([40.85281952, 40.80583952, 41.17077814, ..., 41.2113266 ,
             41.21994522, 41.1978302 ])

    >>> inter_bc_qty(s.stars['age'], s.stars['metallicity'], qty='Vmag', IMF='Chabrier')
    UnitArr([5.5082267 , 5.47273851, 6.07288295, ..., 6.03545475, 5.89715848,
             6.07693768])
    >>> inter_bc_qty(s.stars['age'], s.stars['metallicity'], qty='Vmag', IMF='Salpeter')
    UnitArr([5.86339955, 5.81575021, 6.39816042, ..., 6.29992533, 6.1988034 ,
             6.33598423])

    magnitudes in the SSP tables are normalized to 1 solar mass -
    `derive_rules.get_luminosities` takes care for scaling correctly:
    >>> from ..snapshot import get_luminosities
    >>> L = get_luminosities(s.stars, band='V')
    load block mass... done.
    >>> L
    UnitArr([227724.53566959, 217788.83689963, 126963.87412006, ...,
             126999.67565879, 142102.98979917, 122512.21536806], units="Lsol")

    # To check
    #UnitArr([ 255272.44859605,  236937.66146024,  141531.12688612, ...,
    #          101504.63406881,  143090.23019079,   97253.25881182], units="Lsol")
    >>> lum_to_mag(L)
    UnitArr([-8.64352456, -8.59508904, -8.00920041, ..., -8.00950653,
             -8.13150804, -7.97044848], units="mag")
'''
__all__ = ['load_table', 'inter_bc_qty', 'lum_to_mag']

import numpy as np
from .. import gadget
from ..units import UnitQty, UnitScalar
from ..physics import solar
from .. import environment
import sys

metallicity_name = {
        1e-4: 'm22',
        4e-4: 'm32',
        4e-3: 'm42',
        8e-3: 'm52',
        2e-2: 'm62',
        5e-2: 'm72',
        }

table_names = {
        '1ABmag': ['log-age-yr', 'Mbol', 'g_AB', '(u-g)AB', '(g-r)AB',
                   '(g-i)AB', '(g-z)AB', '(FUV-NUV)AB', '(FUV-r)AB', '(FUV-R)AB',
                   'F(1500A)'],
        '1color': ['log-age-yr', 'Mbol', 'Umag', 'Bmag', 'Vmag', 'Kmag', '14-V',
                   '17-V', '22-V', '27-V', 'U-J', 'J-F', 'F-N', 'U-B', 'B-V'],
        '2color': ['log-age-yr', 'Rmag', 'J2Mmag', 'Kmag', 'V-R', 'V-I', 'V-J',
                   'V-K', 'R-I', 'J-H', 'H-K', 'V-K\'', 'V-Ks', '(J-H)2M ',
                   '(J-Ks)2M'],
        '3color': ['log-age-yr', 'B(4000)', 'B4_VN', 'B4_SDSS', 'B(912)', 'NLy',
                   'NHeI', 'NHeII', 'Mbol', 'Bol_Flux', 'SNR/yr/Lo', 'N(BH)',
                   'N(NS)', 'PNBR/yr/Lo', 'N(WD)', 'M(Remnants)'],
        '4color': ['log-age-yr', 'Mbol', 'Bmag', 'Vmag', 'Kmag', 'M*_liv',
                   'M_remnants', 'M_ret_gas', 'M_galaxy', 'SFR/yr',
                   'M*_liv+M_rem', 'M*_tot/Lb', 'M*_tot/Lv', 'M*_tot/Lk',
                   'M*_liv/Lb', 'M*_liv/Lv', 'M*_liv/Lk'],
        '5color': ['log-age-yr', 'Mbol', 'b(t)*\'s/yr', 'B(t)/yr/Lo',
                   'Turnoff_mass', 'BPMS/BMS'],
        '6lsindx_ffn': ['log-age', 'CN_1', 'CN_2', 'Ca4227', 'G4300', 'Fe4383',
                        'Ca4455', 'Fe4531', 'Fe4668', 'H\\beta', 'Fe5015', 'Mg_1',
                        'Mg_2', 'Mg-b'],
        '6lsindx_sed': ['log-age', 'CN_1', 'CN_2', 'Ca4227', 'G4300', 'Fe4383',
                        'Ca4455', 'Fe4531', 'Fe4668', 'H\\beta', 'Fe5015', 'Mg_1',
                        'Mg_2', 'Mg-b'],
        '6lsindx_sed_lick_system': ['log-age', 'CN_1', 'CN_2', 'Ca4227', 'G4300',
                                    'Fe4383', 'Ca4455', 'Fe4531', 'Fe4668',
                                    'H\\beta', 'Fe5015', 'Mg_1', 'Mg_2', 'Mg-b'],
        '7lsindx_ffn': ['log-age', 'Fe5270', 'Fe5335', 'Fe5406', 'Fe5709',
                        'Fe5782', 'Na-D', 'TiO_1', 'TiO_2', 'H\\delta_A',
                        'H\\gamma_A', 'H\\delta_F', 'H\\gamma_F', 'D(4000)'],
        '7lsindx_sed': ['log-age', 'Fe5270', 'Fe5335', 'Fe5406', 'Fe5709',
                        'Fe5782', 'Na-D', 'TiO_1', 'TiO_2', 'H\\delta_A',
                        'H\\gamma_A', 'H\\delta_F', 'H\\gamma_F', 'D(4000)',
                        'B4_VN', 'CaII8498', 'CaII8542', 'CaII8662', 'MgI8807',
                        'H8_3889', 'H9_3835', 'H10_3798', 'BH-HK'],
        '7lsindx_sed_lick_system': ['log-age', 'Fe5270', 'Fe5335', 'Fe5406',
                                    'Fe5709', 'Fe5782', 'Na-D', 'TiO_1', 'TiO_2',
                                    'H\\delta_A', 'H\\gamma_A', 'H\\delta_F',
                                    'H\\gamma_F', 'D(4000)', 'B4_VN', 'CaII8498',
                                    'CaII8542', 'CaII8662', 'MgI8807', 'H8_3889',
                                    'H9_3835', 'H10_3798', 'BH-HK'],
        '8lsindx_sed_fluxes': ['log-age', 'Nx', 'Im', 'Flux_Blue', 'Flux_Red',
                               'Flux_Ctrl', 'Flux_Line', 'Index'],
        '9color': ['log-age-yr', 'Kmag', 'K-I3.5', 'I3.5-I4.5', 'I4.5-I5.7',
                   'I5.7-I7.9', 'I7.9-I12', 'I12-I25', 'I25-I60', 'I60-I100',
                   'M24-M70', 'M70-M160', 'I100-M160'],
        'acs_wfc_color': ['log-age-yr', 'Vmag', 'Kmag', 'V-F220w', 'V-F250w',
                          'V-F330w', 'V-F410w', 'V-F435w', 'V-F475w', 'V-F555w',
                          'V-F606w', 'V-F625w', 'V-F775w', 'V-F814w', 'log Nly',
                          'Mt/Lb', 'Mt/Lv', 'Mt/Lk'],
        #'ised' -> binary file,
        'wfc3_color': ['log-age-yr', 'Vmag', 'Kmag', 'V-F110W', 'V-F125W',
                       'V-F160W', 'V-F225W', 'V-F336W', 'V-FR388N', 'V-F438W',
                       'V-F555W', 'V-F814W', 'V-ACS220W', 'V-ACS625W', 'log Nly',
                       'Mt/Lb', 'Mt/Lv', 'Mt/Lk'],
        'wfc3_uvis1_color': ['log-age-yr', 'Vmag', 'Kmag', 'V-F225w', 'V-F336w',
                             'V-F438w', 'V-F547m', 'V-F555w', 'V-F606w',
                             'V-F625w', 'V-F656n', 'V-F657n', 'V-F658n',
                             'V-F814w', 'log Nly', 'Mt/Lb', 'Mt/Lv', 'Mt/Lk'],
        'wfpc2_johnson_color': ['log-age-yr', 'Vmag', 'V-F300w', 'V-F300rl',
                                'V-F336w', 'V-F439w', 'V-F450w', 'V-F555w',
                                'V-F606w', 'V-F675w', 'V-F814w', 'V-U', 'V-B',
                                'V-R', 'V-I', 'V-J', 'V-K', 'V-L'],
    }
available_qty = set(np.concatenate(list(table_names.values()))) - set(['log-age-yr', 'log-age'])

table_base_name = {
        'Kroupa': 'bc2003_hr_stelib_%s_kroup_ssp',
        'Chabrier': 'bc2003_hr_stelib_%s_chab_ssp',
        'Salpeter': 'bc2003_hr_stelib_%s_salp_ssp',
        }

def load_table(qty, metal_code, folder=None, base=None, ending=None, IMF=None):
    '''
    Read the column of a quantity from the SSP tables.

    The table path is constructed as in:
        folder + '/' + (base%metal_code) + '.' + ending

    Args:
        qty (str):          The name of the quantity as in the tables.
        metal_code (str):   The part of the file name that encodes the metallicity
                            (values are stored in `metallicity_name`).
        folder (str):       The folder where the SSP model tables are located.
                            Default: as given in the `gadget.cfg` plus a subfolder
                            called as the IMF from `gadget.cfg` in lower cases.
        base (str):         The base name of the tables. Shall include exactly one
                            "%s" which is going to be substituted by the metal
                            code. Default: the string in `table_base_name` of the
                            IMF defined in the `gadget.cfg`.
        ending (str):       The file name ending. If None, it iterates over the
                            keys in `table_names` until the quantity `qty` is
                            found in the table.
        IMF (str):          The name of the IMF to use, if folder and base is not
                            given (undefined behaviour otherwise). By default the
                            one given in `gadget.cfg` is used. Values that are not
                            in `table_base_name.keys()` are invalid.

    Returns:
        age_bins (np.ndarray):  The first column, which are the age bins for the
                                different bins.
        qty (np.ndarray):       The column of the quantity asked for.

    Raises:
        IOError:            If any of the tables does not exist.
        RuntimeError:       If the quantity does not exists in one of the tables.
    '''
    if IMF is None:
        IMF=gadget.general['IMF']
    if folder is None:
        folder = gadget.general['SSP_dir'] + '/' + IMF.lower()
    if base is None:
        base = table_base_name[IMF]

    found = False
    if ending is None:
        for ending, columns in table_names.items():
            if qty in columns:
                tbl = np.loadtxt(folder + '/' + (base%metal_code) + '.' + ending)
                found = True
                break
    else:
        columns = table_names[ending]
        if qty in columns:
            tbl = np.loadtxt(folder + '/' + (base%metal_code) + '.' + ending)
            found = True

    if not found:
        raise RuntimeError('Could not find the quantity "%s" in the ' % qty + \
                           'SSP tables.')
    else:
        if environment.verbose >= environment.VERBOSE_TALKY:
            print("Using SSP table: " + folder + '/' + (base%metal_code) + '.' + ending)


    return tbl[:,0], tbl[:,columns.index(qty)]

def inter_bc_qty(age, Z, qty, units=None, IMF=None):
    '''
    Interpolate a quantity for given age and metallicity from the SSP model
    tables.

    Args:
        age (UnitQty):  The ages. If no units are given, they are assumed to be in
                        years.
        Z (UnitQty):    The metallicities (in absolute units, not solar units).
        qty (str):      The quantity to interpolate (as in the tables).
        IMF (str):      The name of the IMF to use, if folder and base is not
                        given (undefined behaviour otherwise). By default the one
                        given in `gadget.cfg` is used. Values that are not in
                        `table_base_name.keys()` are invalid.

    Returns:
        Q (UnitArr):    The interpolated quantity.

    Raises:
        IOError:            If any of the tables does not exist.
        RuntimeError:       If the quantity does not exists in one of the tables.
        ValueError:         If the length of the ages and the metallicity Z do not
                            fit eachother.
    '''
    if environment.verbose >= environment.VERBOSE_TALKY:
        print('interpolate SSP tables for qty "%s"...' % qty)

    age = np.log10(UnitQty(age, 'yr')).view(np.ndarray)
    Z = UnitQty(Z).view(np.ndarray)

    # allow for single values
    single_value = False
    if not age.shape or not Z.shape:
        single_value = True
        if not age.shape:
            age = np.array([float(age)])
        if not Z.shape:
            Z = np.array([float(Z)])
    if not len(age)==len(Z):
        raise ValueError('The array of ages and metallicities need to have ' + \
                         'the length!')

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('read SSP tables...')
        sys.stdout.flush()
    # load all tables (just a few MB)
    tbls = {}
    for mtl in list(metallicity_name.keys()):
        age_bins, tbls[mtl] = load_table(qty, metallicity_name[mtl], IMF=IMF)

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('table limits:')
        print('  age [yr]:    %.2e - %.2e' % (10**age_bins.min(),
                                              10**age_bins.max()))
        print('  metallicity: %.2e - %.2e' % (min(metallicity_name.keys()),
                                              max(metallicity_name.keys())))

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('interpolate in age...')
        sys.stdout.flush()
    # interpolate in age in all metallicities simultaneously (creating blocks for
    # each tabled metallicity for each particle, i.e. len(metallicity_name) entire
    # blocks -- there are much less metallcity bins than age bins!):
    Q_mtl = {mtl:np.empty(len(age)) for mtl in metallicity_name.keys()}
    for k in range(len(age_bins)+1):
        if k == 0:
            age_mask = age<age_bins[0]
            n = [0, 0]
            a_age = np.zeros(np.sum(age_mask))
        elif k == len(age_bins):
            age_mask = age_bins[-1]<=age
            n = [-1, -1]
            a_age = np.zeros(np.sum(age_mask))
        else:
            age_mask = (age_bins[k-1]<=age) & (age<age_bins[k])
            n = [k-1, k]
            a_age = (age[age_mask] - age_bins[n[0]]) / (age_bins[n[1]] - age_bins[n[0]])

        for mtl in metallicity_name.keys():
            tbl = tbls[mtl]
            Q_mtl[mtl][age_mask] = a_age*tbl[n[1]] + (1.0-a_age)*tbl[n[0]]

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('interpolate in metallicity...')
        sys.stdout.flush()
    # now interpolate in metallicity:
    Q = np.empty(len(age))
    for i in range(len(metallicity_name)+1):
        if i == 0:
            z = list(metallicity_name.keys())[0]
            Z_mask = Z<z
            Q[Z_mask] = Q_mtl[z][Z_mask]
        elif i == len(metallicity_name):
            z = list(metallicity_name.keys())[-1]
            Z_mask = z<=Z
            Q[Z_mask] = Q_mtl[z][Z_mask]
        else:
            z = [list(metallicity_name.keys())[i-1], list(metallicity_name.keys())[i]]
            Z_mask = (z[0]<=Z) & (Z<z[1])
            a_Z = (Z[Z_mask] - z[0]) / (z[1] - z[0])
            Q[Z_mask] = a_Z * Q_mtl[z[1]][Z_mask] + \
                    (1.0-a_Z) * Q_mtl[z[0]][Z_mask]

    if single_value:
        return UnitScalar(float(Q), units)
    return UnitQty(Q, units)

def lum_to_mag(L, subs=None):
    '''
    Convert luminosities to magnitudes.

    Args:
        L (UnitQty):    The luminosities to convert (if not units are given, they
                        are assumed to be in 'Lsol').

    Returns:
        mag (UnitArr):  The magnitudes.
    '''
    L = UnitQty(L, 'Lsol', subs=subs, dtype=float)
    return UnitQty( solar.abs_mag - 2.5 * np.log10(L), 'mag')




'''
A collection of (sub-)snapshot wide analysis functions.

Example:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_320', physical=False)
    >>> if np.linalg.norm(mass_weighted_mean(s,'pos') - center_of_mass(s)) > 1e-3:
    ...     print(mass_weighted_mean(s,'pos'))
    ...     print(center_of_mass(s,))
    load block pos... done.
    load block mass... done.

    Center of mass is *not* center of galaxy / halo!
    >>> Translation([-34792.2, -35584.8, -33617.9]).apply(s)
    apply Translation to "pos" of "snap_M1196_4x_320"... done.
    >>> s.to_physical_units()
    >>> sub = s[s['r'] < '20 kpc']
    derive block r... done.
    >>> orientate_at(sub, 'L', total=True)
    load block vel... done.
    derive block momentum... done.
    derive block angmom... done.
    apply Rotation to "pos" of "snap_M1196_4x_320"... done.
    apply Rotation to "vel" of "snap_M1196_4x_320"... done.
    >>> if np.linalg.norm(sub['angmom'].sum(axis=0) -
    ...         UnitArr([3.18e+08,-5.05e+07,1.24e+14],'kpc Msol km/s')) > 1e12:
    ...     print(sub['angmom'].sum(axis=0))
    derive block momentum... done.
    derive block angmom... done.
    >>> redI = reduced_inertia_tensor(sub.baryons)
    derive block r... done.
    >>> np.linalg.norm(redI - np.matrix([[ 2.54e10, -3.28e9, 2.74e9],
    ...                                  [-3.28e9,   2.75e9, 5.16e8],
    ...                                  [ 2.74e9,   5.16e8, 5.84e9]])) < 1e10
    True
    >>> orientate_at(s, 'red I', redI)
    apply Rotation to "pos" of "snap_M1196_4x_320"... done.
    apply Rotation to "vel" of "snap_M1196_4x_320"... done.
    >>> redI = reduced_inertia_tensor(sub.baryons)
    derive block r... done.
    >>> np.linalg.norm(redI - np.matrix([[ 2.62e10,  0.00   , 0.00   ],
    ...                                  [ 0.00   ,  5.73e+9, 0.00   ],
    ...                                  [ 0.00   ,  0.00   , 2.05e9]])) < 1e8
    True
    >>> if abs( los_velocity_dispersion(sub) - '170 km/s' ) > '5 km/s':
    ...     print(round(float(los_velocity_dispersion(sub)),4))
    119.8176
    >>> if abs(half_mass_radius(sub.stars) - '11 kpc') > '0.8 kpc':
    ...     print(half_mass_radius(sub.stars))
    >>> if abs(eff_radius(sub, 'V', proj=None) - '11 kpc') > '0.8 kpc':
    ...     print(eff_radius(sub, 'V', proj=None))
    load block form_time... done.
    derive block age... done.
    load block Z... done.
    derive block elements... done.
    derive block H... done.
    derive block He... done.
    derive block metals... done.
    derive block metallicity... done.
    derive block lum_v... done.
    >>> if abs(eff_radius(sub, 'V', proj=2) - '10.7 kpc') > '0.2 kpc':
    ...     print(eff_radius(sub, 'V', proj=2))
    derive block rcyl... done.
    >>> if abs(half_mass_radius(sub.stars, proj=2) - '10.8 kpc') > '0.2 kpc':
    ...     print(half_mass_radius(sub.stars))
    >>> ifr, ofr = flow_rates(s, '50 kpc')
    derive block vrad... done.
    >>> if abs(ifr - '355 Msol/yr') > '10 Msol/yr' or abs(ofr - '337 Msol/yr') > '10 Msol/yr':
    ...     print(ifr, ofr)
    >>> if abs(shell_flow_rates(s.gas, UnitArr([48,52],'kpc')) - '-4.6 Msol/yr') > '0.2 Msol/yr':
    ...     print(shell_flow_rates(s.gas, UnitArr([48,52],'kpc')))
    >>> if abs(shell_flow_rates(s.gas, UnitArr([48,52],'kpc'), qty='metals') - '-0.014 Msol/yr') > '0.002 Msol/yr':
    ...     print(shell_flow_rates(s.gas, UnitArr([48,52],'kpc'), qty='metals'))
    >>> if abs(shell_flow_rates(s.gas, UnitArr([48,52],'kpc'), direction='in') - '-12.2 Msol/yr') > '0.2 Msol/yr':
    ...     print(shell_flow_rates(s.gas, UnitArr([48,52],'kpc'), direction='in'))
    >>> if abs(shell_flow_rates(s.gas, UnitArr([48,52],'kpc'), direction='out') - '7.6 Msol/yr') > '0.2 Msol/yr':
    ...     print(shell_flow_rates(s.gas, UnitArr([48,52],'kpc'), direction='out'))
    >>> ifr, ofr = flow_rates(s.gas, '50 kpc')

    >>> if abs(ifr - '13.0 Msol/yr') > '1.0 Msol/yr' or abs(ofr - '9.1 Msol/yr') > '0.1 Msol/yr':
    ...     print(ifr, ofr)
    >>> eta = ofr / s.gas['sfr'].sum()
    load block sfr... done.
    >>> if abs(eta - 2.0) > 0.2:
    ...     print('mass loading:', eta)

    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', load_double_prec=True)
    >>> s.gas['lx'] = x_ray_luminosity(s, lumtable=module_dir+'snaps/em.dat')
    load block Z... done.
    derive block elements... done.
    derive block H... done.
    derive block He... done.
    derive block metals... done.
    derive block metallicity... done.
    load block ne... done.
    load block rho... done.
    load block mass... done.
    load block u... done.
    derive block temp... done.
    >>> s.gas['lx'].units
    Unit("erg s**-1")
    >>> if abs(np.mean(s.gas['lx']) - '1.68e33 erg/s') > '0.05e33 erg/s':
    ...     print(np.mean(s.gas['lx']))
    >>> if abs(s.gas['lx'].sum() - '1.55e39 erg/s') > '0.05e39 erg/s':
    ...     print(s.gas['lx'].sum())
'''
__all__ = ['mass_weighted_mean', 'center_of_mass', 'reduced_inertia_tensor',
           'orientate_at', 'half_qty_radius', 'half_mass_radius', 'eff_radius',
           'shell_flow_rates', 'flow_rates', 'los_velocity_dispersion',
           'x_ray_luminosity']

import numpy as np
from ..units import *
from ..utils import dist
from ..transformation import *
from pygad import physics
import sys


def mass_weighted_mean(s, qty, mass='mass'):
    '''
    Calculate the mass weighted mean of some quantity, i.e.
    sum(mass[i]*qty[i])/sum(mass).

    Args:
        s (Snap):           The (sub-)snapshot the quantity belongs to (the masses
                            are taken from it, too).
        qty (str, UnitArr): The quantity to average. It can either be the block
                            itself (make shure it already has the appropiate
                            shape), or a expression that can be passed to Snap.get
                            (e.g. simply a name of a block).
        mass (str, UnitArr):The mass block.

    Returns:
        mean (UnitArr):     The mass weighted mean.
    '''
    if isinstance(qty, str):
        qty = s.get(qty)
    else:
        qty = UnitArr(qty)
    if len(s) == 0:
        return UnitArr([0] * qty.shape[-1], units=qty.units, dtype=qty.dtype)
    if isinstance(mass, str):
        mass = s.get(mass)
    else:
        mass = UnitArr(mass)
    # only using the np.ndarray views does not speed up
    mwgt = np.tensordot(mass, qty, axes=1)
    normalized_mwgt = mwgt / float(mass.sum())
    return UnitArr(normalized_mwgt, qty.units)


def center_of_mass(snap):
    '''Calculate and return the center of mass of this snapshot.'''
    return mass_weighted_mean(snap, 'pos')


def reduced_inertia_tensor(s):
    '''
    Calculate the 'reduced' inertia tensor by Gerhard (1983) / Bailin & Steinmetz
    (2005) of this ensemble.

    $I_ij = \\sum_k m_k \\frac{r_{k,i} r_{k,j}}{r_k^2}$
    I_ij = sum_k m_k (r_ki r_kj) / r_k^2

    Args:
        s (Snap):   The (sub-)snapshot to calculate the reduced inertia tensor of.

    Returns:
        I (np.matrix):  The reduced inertia tensor. (Without units, but they would
                        be s['mass'].units/s['pos'].units.)
    '''
    # a bit faster with the np.ndarray views
    r2 = (s['r'] ** 2).view(np.ndarray)
    m = s['mass'].view(np.ndarray)
    pos = s['pos'].view(np.ndarray)

    # do not divide by zero
    rzero = (r2 == 0)
    if np.any(rzero):
        rzero = ~rzero
        r2 = r2[rzero]
        m = m[rzero]
        pos = pos[rzero]

    I_xx = np.sum(m * pos[:, 0] ** 2 / r2)
    I_yy = np.sum(m * pos[:, 1] ** 2 / r2)
    I_zz = np.sum(m * pos[:, 2] ** 2 / r2)
    I_xy = np.sum(m * pos[:, 0] * pos[:, 1] / r2)
    I_xz = np.sum(m * pos[:, 0] * pos[:, 2] / r2)
    I_yz = np.sum(m * pos[:, 1] * pos[:, 2] / r2)
    I = np.matrix([[I_xx, I_xy, I_xz], \
                   [I_xy, I_yy, I_yz], \
                   [I_xz, I_yz, I_zz]], dtype=np.float64)
    return I


def orientate_at(s, mode, qty=None, total=False, remember=True):
    '''
    Orientate the (sub-)snapshot at a given quantity.

    Possible modes:
        'vec'/'L':          Orientate such that the given vector alignes with the
                            z-axis. If no vector is given and the mode is 'L', the
                            angular momentum is used.
        'tensor'/'red I':   Orientate at the eigenvectors of a tensor. The
                            eigenvector with the smallest eigenvalue is
                            orientated along the z-axis and the one with the
                            largest eigenvalue along the x-axis. (Hence, the
                            orientation to 'red I' is similar to the orientation
                            to 'L'). If no tensor is given and the mode is
                            'red I', the reduced inertia tensor is used.

    Args:
        s (Snap):       The snapshot to orientate
        mode (str):     The mode of orientation. See above.
        qty (...):      The quantity to use for orientation. If it is None, it is
                        calculated on the fly (for the passed (sub-)snapshot).
        total (bool):   Whether to apply the transformation to the entire snapshot
                        or just the passed sub-snapshot. (Cf. Transformation.apply
                        for more information!)
        remember (bool):
                        Remember the transformation for blocks loaded later. (Cf.
                        Transformation.apply for more information!)
    '''
    if qty is None:
        if mode == 'L':
            qty = s['angmom'].sum(axis=0)
        elif mode == 'red I':
            qty = reduced_inertia_tensor(s)
        else:
            raise ValueError('No quantity passed to orientate at!')

    if mode in ['vec', 'L']:
        T = rot_to_z(qty)
    elif mode in ['tensor', 'red I']:
        qty = np.matrix(qty)
        if np.max(np.abs(qty.H - qty)) > 1e-6:
            raise ValueError('The matrix passed as qty has to be Hermitian!')
        vals, vecs = np.linalg.eigh(qty)
        i = np.argsort(vals)[::-1]
        try:
            T = Rotation(vecs[:, i].T)
        except ValueError:
            # probably not a proper rotation... (not right-handed)
            vecs[:, i[1]] *= -1
            T = Rotation(vecs[:, i].T)
        except:
            raise
    else:
        raise ValueError('unknown orientation mode \'%s\'' % what)

    T.apply(s, total=total, remember=remember)


def half_qty_radius(s, qty, Qtot=None, center=None, proj=None):
    '''
    Calculate the radius at which half of a quantity is confined in.

    Args:
        s (Snap):               The (sub-)snapshot to use.
        qty (str, UnitQty):     The quantity to calculate the half-X-radius for.
        Qtot (float, UnitArr):  The total amount of the quantity. If not given,
                                the total quantity of the (sub-)snapshot is used.
        center (array-like):    The center of the structure to calculate the
                                properties for. (default: [0,0,0])
        proj (int):             If set, do the calculation for the projection
                                along the specified axis (0=x, 1=y, 2=z).

    Returns:
        r12 (UnitArr):          The half-X-radius.
    '''
    if len(s) == 0:
        # leads to exceptions otherwise
        return UnitArr(0.0, s['pos'].units)

    if center is None:
        center = UnitQty([0] * 3)
    center = UnitQty(center, s['pos'].units, subs=s)

    if isinstance(proj, int):
        proj_mask = [i for i in range(3) if i != proj]
        if len(center) == 3:
            center = center[proj_mask]

    if np.all(center == 0):
        if isinstance(proj, int):
            r = s['rcyl'] if proj == 2 else dist(s['pos'][:, proj_mask])
        else:
            r = s['r']
    else:
        r = dist(s['pos'][:, proj_mask], center) if isinstance(proj, int) \
            else dist(s['pos'], center)
    r_ind = r.argsort()

    if isinstance(qty, str):
        qty = s.get(qty)
    else:
        qty = UnitQty(qty)

    Q = np.cumsum(qty[r_ind])
    if Qtot is None:
        Qtot = Q[-1]
    else:
        Qtot = UnitScalar(Qtot, qty.units, subs=s, dtype=float)

    Q_half_ind = np.abs(Q - Qtot / 2.).argmin()
    if Q_half_ind == len(Q) - 1:
        print('WARNING: The half-qty radius is larger than ' + \
              'the (sub-)snapshot passed!', file=sys.stderr)
    elif Q_half_ind < 10:
        print('WARNING: The half-qty radius is not resolved ' + \
              'for %s!' % s, file=sys.stderr)
    return UnitArr(r[r_ind][Q_half_ind], s['pos'].units)


def half_mass_radius(s, M=None, center=None, proj=None):
    '''
    Calculate the (by default 3D) half-mass-radius of the structure at the center.

    For more details see analysis.half_qty_radius.

    Args:
        s (Snap):               The (sub-)snapshot to use.
        M (float, UnitArr):     The total mass.
        center (array-like):    The center of the structure. (default: [0,0,0])
        proj (int):             If set, do the calculation for the projection
                                along the specified axis (0=x, 1=y, 2=z).

    Returns:
        r12 (UnitArr):          The half-mass-radius.
    '''
    return half_qty_radius(s, qty='mass', Qtot=M, center=center, proj=proj)


def eff_radius(s, band=None, L=None, center=None, proj=2):
    '''
    Calculate the (by default 2D) half-light-radius of the structure at the
    center.

    For more details see analysis.half_qty_radius.

    Args:
        s (Snap):               The (sub-)snapshot to use.
        band (str):             The band in which the effective radius is
                                calculated. If it is None or 'bol', the bolometric
                                luminosity is taken.
        L (float, UnitArr):     The total luminosity.
        center (array-like):    The center of the structure. (default: [0,0,0])
        proj (int):             The axis to project along.

    Returns:
        r12 (UnitArr):          The half-light-radius.
    '''
    qty = 'lum'
    if band is not None and band != 'bol':
        qty += '_' + band.lower()
    return half_qty_radius(s.stars, qty=qty, Qtot=L, center=center, proj=proj)


def shell_flow_rates(s, Rlim, qty='mass', direction='both', units='Msol/yr'):
    '''
    Estimate flow rate in spherical shell.

    The estimation is done by caluculating m*v/d on a particle base, where d is
    the thickness of the shell.

    Args:
        s (Snap):               The (sub-)snapshot to use.
        Rlim (UnitQty):         The inner and outer radius of the shell.
        qty (str, UnitQty):     The (mass) quantity to calculate the flow rates of.
        direction ('in', 'out', 'both'):
                                The direction to take into account. If 'in', only
                                take onflowing gas particles into the calculation;
                                analog for 'out'; and do not restrict to particles
                                with 'both'.
        units (Unit, str):      The units in which to return the flow rate.

    Returns:
        flow (UnitArr):         The estimated flow rate.
    '''
    Rlim = UnitQty(Rlim, s['r'].units)
    if not Rlim.shape == (2,):
        raise ValueError('Rlim must have shape (2,)!')
    shell = s[(Rlim[0] < s['r']) & (s['r'] < Rlim[1])]

    if direction == 'both':
        pass
    elif direction == 'in':
        shell = shell[shell['vrad'] < 0]
    elif direction == 'out':
        shell = shell[shell['vrad'] > 0]
    else:
        raise RuntimeError('unknown direction %s!' % direction)

    if isinstance(qty, str):
        qty = shell.get(qty)
    else:
        qty = UnitArr(qty, 'Msol', subs=s)

    flow = np.sum(qty * shell['vrad'] / UnitArr(Rlim[1] - Rlim[0], Rlim.units))
    flow.convert_to(units, subs=s)
    return flow


def flow_rates(s, R, qty='mass', dt='3 Myr'):
    '''
    Estimate in- and outflow rates of a given quantity (default: mass)
    through a given radius.

    The estimation is done by propagating the positions with constant current
    velocities, i.e. pos_new = pos_old + vel*dt. Then counting the mass that
    passed the shell of radius R.

    Args:
        s (Snap):               The (sub-)snapshot to use.
        qty (str):              The quantity to calculate the flow-rates of
                                (default: 'mass')
        R (UnitScalar):         The radius of the shell through which the flow is
                                estimated.
        dt (UnitScalar):        The time used for the linear extrapolation of the
                                current positions (if it is a plain number without
                                units, they are assumed to be 'Myr').

    Returns:
        ifr (UnitArr):          The estimated inflow rate.
        ofr (UnitArr):          The estimated outflow rate.
    '''
    R = UnitScalar(R, units=s['pos'].units, subs=s)
    dt = UnitScalar(dt, units='Myr', subs=s)

    # predicted particle distances from center in dt
    dt.convert_to(s['r'].units / s['vel'].units, subs=s)  # avoid conversions of
    # entire arrays
    rpred = s['r'] + s['vrad'] * dt
    of_mass = s[qty][(s['r'] < R) & (rpred >= R)]
    if_mass = s[qty][(s['r'] >= R) & (rpred < R)]

    dt.convert_to('yr', subs=s)  # to more intuitive units again
    ofr = np.sum(of_mass) / dt
    ifr = np.sum(if_mass) / dt

    return ifr, ofr


def los_velocity_dispersion(s, proj=2):
    '''
    Calculate (mass-weighted) line-of-sight velocity dispersion.

    Args:
        s (Snap):       The (sub-)snapshot to use.
        proj (int):     The line of sight is along this axis (0=x, 1=y, 2=z).
    '''
    # array of los velocities
    v = s['vel'][:, proj].ravel()
    av_v = mass_weighted_mean(s, v)
    sigma_v = np.sqrt(mass_weighted_mean(s, (v - av_v) ** 2))

    return sigma_v


def x_ray_luminosity(s, lumtable='em.dat', tempbin=None, lx0bin=None, dlxbin=None,
                     Zref=0.4, z_table=0.001):
    '''
    Calculate X-ray luminosity of gas particles using a prepared emission table
    from XSPEC.

    Args:
        s (Snap):               The snapshot to use.
        lumtable (str):         The filename of the XSPEC emission table to use
                                (default: 'em.dat' in current directory)
        tempbin, lx0bin, dlxbin (array-like):
                                Temperature, Lx0 and dLx bins: Can be passed
                                instead of lumtable file (e.g. to avoid reading
                                in the same file multiple times in a loop)
        Zref (float):           The reference metallicity used for the XSPEC
                                table.
        z_table (float):        The redshift assumed for the XSPEC table.

    Returns:
        lx (UnitArr):           X-ray luminosities of the gas particles
    '''
    if abs(s.redshift - z_table) > 1e-2:
        # from pygad-dsorini
        print('WARNING: Snapshot\'s redshift (%.3g) does not ' % s.redshift +
                           'match the table\'s redshift (%.3g)!' % z_table)
        # raise RuntimeError('Snapshot\'s redshift (%.3g) does not ' % s.redshift +
        #                    'match the table\'s redshift (%.3g)!' % z_table)

    # Read in temperature bins and corresponding Lx0(T,Z=Zref) and (dLx/dZ)(T)
    # (both in 1e44 erg/s (per Zsol))
    if tempbin == None:
        tempbin, lx0bin, dlxbin = np.loadtxt(lumtable, usecols=(0, 3, 5), unpack=True)
    else:
        tempbin = np.asarray(tempbin)

    tlow = tempbin[0] - 0.5 * (tempbin[1] - tempbin[0])  # lower temperature bin limit
    Z = s.gas['metallicity'] / physics.solar.Z()  # metallicity in solar units
    mp = physics.m_p.in_units_of('g')  # proton mass

    # Horst: configurability of H-property
    if s.H_neutral_only:
        # Romeel 28/5/2018: can't use s.gas['H'] in our Gizmo sims, this corresponds to *neutral* H; have to get H mass from total-helium-metals
        mhydr = np.float64(s.gas['mass']).in_units_of('g')*(1.-Z*physics.solar.Z())-np.float64(s.gas['He']).in_units_of('g') # H mass in g
    else:
        mhydr = np.float64(s.gas['H']).in_units_of('g')

    #print (1.-Z*physics.solar.Z()),'H in g',mhydr[Z>0]
    # emission measure of gas particles (n_e * n_H * V)
    # Horst: configurability of H-property
    #em = np.float64(s.gas['ne']) * np.float64(s.gas['H']).in_units_of('g') ** 2 * \
    em = np.float64(s.gas['ne']) * mhydr**2 * \
         np.float64(s.gas['rho']).in_units_of('g/cm**3') / \
         (np.float64(s.gas['mass']).in_units_of('g') * mp ** 2)
    # rescaling factor for precomputed luminosities
    Da = s.cosmology.angular_diameter_distance(z_table, 'cm')
    norm = UnitArr(1e-14, units='cm**5') * em / (4 * np.pi * (Da * (1 + z_table)) ** 2)
    norm = norm.view(np.ndarray)
    lx = np.zeros(s.gas['rho'].shape[0])  # array for X-ray luminosity
    kB_T = (s.gas['temp'] * physics.kB).in_units_of('keV').view(np.ndarray)
    indices = np.zeros(s.gas['rho'].shape[0])  # array for fitting tempbin
    # indices for gas particles
    dtemp = np.zeros(s.gas['rho'].shape[0]) + 1e30  # minimal differences of gas
    # particle temperature and binned temperatures

    # loop over tempbin array to find nearest tempbin for all gas particle
    # temperatures
    for i in range(0, tempbin.shape[0]):
        dtemp[np.where(np.abs(kB_T - tempbin[i]) < dtemp)] = \
            np.abs(kB_T[np.where(np.abs(kB_T - tempbin[i]) < dtemp)] - tempbin[i])
        indices[np.where(np.abs(kB_T - tempbin[i]) == dtemp)] = i

    # calculate X-ray luminosities for all gas particles
    for i in range(0, tempbin.shape[0]):
        lx[np.where(indices == i)] = lx0bin[i] + \
                                     (Z[np.where(indices == i)] - Zref) * dlxbin[i]
    lx[np.where(kB_T < tlow)] = 0  # particles below threshold temperature do not
    # contribute to Lx
    return UnitArr(lx * norm * 1e44, 'erg/s')  # luminosities of all gas particles [erg/s]



#!/usr/bin/env python
import os
import subprocess
from glob import glob

from setuptools import Extension, setup

import versioneer

# define scripts
scripts = ["bin/ginsp", "bin/gconv", "bin/gCache3", "bin/gCatalog3", "bin/gStarform3"]

# find all sub-packages
modules = []
setup_dir = os.path.dirname(os.path.realpath(__file__))
for root, dirs, files in os.walk(setup_dir):
    submod = os.path.relpath(root, setup_dir).replace(os.sep, ".")
    if not submod.startswith("pygad"):
        continue
    if "__init__.py" in files:
        modules.append(submod)

# clean and make the cpygad.so library
subprocess.run(["make", "clean"], cwd=setup_dir + "/pygad/C", check=True)

gsl_include = ""
gsl_lib = ""
if os.getenv("GSL_HOME") is not None:
    gsl_include = os.getenv("GSL_HOME") + "/include"
    gsl_lib = os.getenv("GSL_HOME") + "/lib"

ext_module = Extension(
    "pygad/C/cpygad",
    language="c++",
    sources=glob("pygad/C/src/*"),
    include_dirs=[
        "pygad/C/include",
        "/usr/include",
        gsl_include,
    ],
    extra_compile_args=[
        "-fPIC",
        "-std=c++11",
        "-O3",
        "-fopenmp",
        "-pedantic",
        "-Wall",
        "-Wextra",
    ],
    libraries=["m", "gsl", "gslcblas", "gomp"],
    extra_link_args=["-fopenmp"],
    library_dirs=[gsl_lib],
)

setup(
    name="pygadmpa",
    description="analysis module for Gadget",
    long_description="A light-weighted analysis module for galaxy \
        simulations performed by the SPH code Gadget.",
    author="Bernhard Roettgers",
    author_email="broett@mpa-garching.mpg.de",
    url="https://bitbucket.org/broett/pygad",
    include_package_data=True,
    packages=list(map(str, modules)),
    scripts=scripts,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=[ext_module],
)



'''
Definition of more complicated derived blocks. May be changed bu user.

A function for a derived block shall return the derived block with units as a
`UnitArr` and shall have an attribute `_deps` that is a set of the names of its
direct dependencies, i.e. the blocks it needs directly to calculate the derived
one from.
'''
__all__ = ['calc_temps', 'age_from_form', 'calc_x_ray_lum', 'calc_HI_mass',
           'calc_ion_mass', 'calc_cooling_rates', 'get_luminosities']

from .. import environment
from ..units import UnitArr, UnitScalar, UnitQty, UnitError
import numpy as np
from .. import physics
from .. import gadget
from .. import cloudy
from . import derived
from fractions import Fraction
from multiprocessing import Pool, cpu_count
import warnings
import gc
import sys


def calc_cooling_rates(s, tbl='CoolingTables/z_0.000.hdf5'):
    '''
    Calculate the total cooling rates for the given snapshot.
    '''
    tbl = physics.cooling.Wiersma_CoolingTable(tbl)
    Lambda = tbl.get_cooling(s, units='erg cm**3 s**-1')
    return Lambda


# TODO: add the elements individually (not just 'elements')
calc_cooling_rates._deps = set(['elements', 'mass', 'Z', 'temp', 'rho'])


def calc_temps(u, XH=0.76, ne=0.0, XZ=None, f=3, subs=None):
    '''
    Calculate the block of temperatures form internal energy.

    This function calculates the temperatures from the internal energy using the
    ideal gas law,
        U = f/2 N k_b T
    and calculating an average particle mass from XH, ne, and XZ.

    Args:
        u (UnitQty):            The (mass-)specific internal energies.
        XH (float, array-like): The Hydrogen *mass* fraction(s). It can be either
                                a constant value or one for each particle
                                (H/mass).
        ne (float, array-like): The number of electrons per atom(!).
        XZ (iterable):          The metal mass fraction(s) and their atomic weight
                                in atomic units. Each element of this iterable
                                shall be a tuple of mass fraction (which can be an
                                array with fractions for each particle) at first
                                position and atomic weight (a single float) at
                                second postion.
                                The Helium mass fractions is always the remaining
                                mass:  1.0 - XH - sum(XZ[:,0]).
        f (float):              The (effective) degrees of freedom.
        subs (dict, Snap):      Substitutions used for conversion into Kelvin.
                                (c.f. `UnitArr.convert_to`)

    Returns:
        T (UnitArr):            The temperatures for the particles (in K if
                                possible).
    '''
    u = UnitQty(u, 'km**2/s**2', subs=subs)
    XH = np.array(XH)
    ne = np.array(ne)

    tmp = XH / 1.008
    XHe = 1.0 - XH
    if XZ is not None:
        for X, m in XZ:
            tmp += X / float(m)
            XHe -= X
    tmp += XHe / 4.003

    # assuming `ne` is the mean number of electrons per atom:
    # av_m = physics.m_u / (tmp * (1.0 + ne))
    # as in Gadget (where XZ=None, though): `ne` are the electrons per Hydrogen
    # atom:
    av_m = physics.m_u / (tmp + ne * XH)

    # solving:  U = f/2 N k_b T
    # which is:  u = f/2 N/M k_b T
    T = u / (f / 2.) * av_m / physics.kB
    try:
        T.convert_to('K', subs=subs)
    except UnitError as ue:
        print('WARNING: in "calc_temps":\n%s' % ue, file=sys.stderr)
    gc.collect()
    return T


# (additional) dependencies
calc_temps._deps = set()


def _z2Gyr_vec(arr, cosmo):
    '''Needed to pickle cosmo.lookback_time_in_Gyr for Pool().apply_async.'''
    return np.vectorize(cosmo.lookback_time_in_Gyr)(arr)


def age_from_form(form, subs, cosmic_time=None, cosmo=None, units='Gyr', parallel=None):
    '''
    Calculate ages from formation time.

    Args:
        form (UnitArr):     The formation times to convert. Has to be UnitArr with
                            appropiate units, i.e. '*_form' or a time unit.
        subs (dict, Snap):  Subsitution for unit convertions. See e.g.
                            `UnitArr.convert_to` for more information.
        cosmic_time (UnitScalar):
                            The current cosmic time. If None and subs is a
                            snapshot, it defaults to subs.time.
        cosmo (FLRWCosmo):  A cosmology to use for conversions. If None and subs
                            is a snapshot, the cosmology of that snapshot is used.
        units (str, Unit):  The units to return the ages in. If None, the return
                            value still has correct units, you just do not have
                            control over them.
        parallel (bool):    If units are converted from Gyr (or some other time
                            unit) to z_form / a_form, one can choose to use
                            multiple threads. By default, the function chooses
                            automatically whether to perform in parallel or not.

    Returns:
        ages (UnitArr):     The ages.

    Examples:
        >>> cosmo = physics.Planck2013()
        >>> age_from_form(UnitArr([0.001, 0.1, 0.5, 0.9], 'a_form'),
        ...               subs={'a':0.9, 'z':physics.a2z(0.9)},
        ...               cosmo=cosmo)
        UnitArr([ 12.30493079,  11.75687265,   6.44278561,   0.        ], units="Gyr")
        >>> age_from_form(UnitArr([10.0, 1.0, 0.5, 0.1], 'z_form'),
        ...               subs={'a':0.9, 'z':physics.a2z(0.9)},
        ...               cosmo=cosmo)
        UnitArr([ 11.82987008,   6.44278561,   3.70734788,  -0.13772577], units="Gyr")
        >>> age_from_form(UnitArr([-2.0, 0.0, 1.0], '(ckpc h_0**-1) / (km/s)'),
        ...               cosmic_time='2.1 Gyr',
        ...               subs={'a':0.9, 'z':physics.a2z(0.9), 'h_0':cosmo.h_0},
        ...               cosmo=cosmo,
        ...               units='Myr')
        UnitArr([ 4697.05769369,  2100.        ,   801.47115316], units="Myr")
    '''
    from ..snapshot.snapshot import Snapshot
    if subs is None:
        subs = {}
    elif isinstance(subs, Snapshot):
        snap, subs = subs, {}
        subs['a'] = snap.scale_factor
        subs['z'] = snap.redshift
        subs['h_0'] = snap.cosmology.h_0
        if cosmo is None:
            cosmo = snap.cosmology
        if cosmic_time is None:
            cosmic_time = UnitScalar(snap.time, gadget.get_block_units('AGE '),
                                     subs=subs)

    form = form.copy().view(UnitArr)

    if str(form.units).endswith('_form]'):
        # (a ->) z -> Gyr (-> time_units)
        if form.units == 'a_form':
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                form.setfield(np.vectorize(physics.a2z)(form), dtype=form.dtype)
        form.units = 'z_form'

        if environment.allow_parallel_conversion and (
                parallel or (parallel is None and len(form) > 1000)):
            N_threads = cpu_count()
            chunk = [[i * len(form) // N_threads, (i + 1) * len(form) // N_threads]
                     for i in range(N_threads)]
            p = Pool(N_threads)
            res = [None] * N_threads
            with warnings.catch_warnings():
                # warnings.catch_warnings doesn't work in parallel
                # environment...
                warnings.simplefilter("ignore")  # for _z2Gyr_vec
                for i in range(N_threads):
                    res[i] = p.apply_async(_z2Gyr_vec,
                                           (form[chunk[i][0]:chunk[i][1]], cosmo))
            for i in range(N_threads):
                form[chunk[i][0]:chunk[i][1]] = res[i].get()
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")  # for _z2Gyr_vec
                form.setfield(_z2Gyr_vec(form, cosmo), dtype=form.dtype)
        form.units = 'Gyr'
        # from present day ages (lookback time) to actual current ages
        form -= cosmo.lookback_time(subs['z'])

    else:
        # 't_form' -> actual age
        cosmic_time = UnitScalar(cosmic_time, form.units, subs=subs)
        form = cosmic_time - form

    if units is not None:
        form.convert_to(units, subs=subs)

    return form


age_from_form._deps = set(['form_time'])


def calc_x_ray_lum(s, lumtable, **kwargs):
    '''
    Wrapping `x_ray_luminosity` for derived blocks.

    Args:
        s (Snap):           The snapshot to use.
        lumtable (str):     The filename of the XSPEC emission table to use.
        **kwargs:           Further arguments passed to `x_ray_luminosity`.

    Returns:
        lx (UnitArr):       X-ray luminosities of the gas particles.
    '''
    from ..analysis import x_ray_luminosity
    return x_ray_luminosity(s, lumtable=lumtable, **kwargs)


calc_x_ray_lum._deps = set(['metallicity', 'ne', 'H', 'rho', 'mass', 'temp'])


def calc_HI_mass(s, UVB=gadget.general['UVB'], flux_factor=None):
    '''
    Estimate the HI mass with the fitting formula from Rahmati et al. (2013).

    It just calls `cloudy.Rahmati_HI_mass`, but this function has `_deps`.

    Args:
        s (Snap):       The (gas-particles sub-)snapshot to use.
        UVB (str):      The name of the UV background as named in `cloudy.UVB`.
                        Defaults to the value of the UVB in the `gadget.cfg`.
        flux_factor (float):
                        Adjust the UVB by this factor (assume a optically thin
                        limit and scale down the densities during the look-up by
                        this factor).
                        (Note: `sigHI` used in the Rahmati fitting formula is not
                        adjusted!)

    Returns:
        HI (UnitArr):   The HI mass block for the gas (within `s`).
    '''
    return cloudy.Rahmati_HI_mass(s, UVB, flux_factor=flux_factor)


calc_HI_mass._deps = set(['H', 'temp', 'rho', 'mass'])


def calc_ion_mass(s, el, ionisation, selfshield=True, iontbl=None,
                  warn_outofbounds=True):
    '''
    Calculate the mass of the given ion from Cloudy tables.

    Args:
        s (Snap):           The (gas-particles sub-)snapshot to use.
        el (str):           The name of the ion-element. Needs also to be a block
                            name (e.g. as one of the 'element' block).
        ionisation (str):   The ionisation state. The concatenation of `el` and
                            this must result in an ion name in the Cloudy table.
        selfshield (bool):  Whether to account for self-shielding by assuming the
                            attenuation of the UVB as experienced by HI using the
                            Rahmati+ (2013) prescription (formula (14) of the
                            paper). Note that this is just a rough approximation!
        iontbl (IonisationTable):
                            The ionisation table to use for the table
                            interpolation. Default to the one given by
                            `derived.cfg` for the redshift of the (sub-)snapshot.
        warn_outofbounds (bool):
                            Whether to print out a warning if particles are
                            out of the bounds of the ionisation tables.

    Returns:
        ion_mass (UnitArr): The mass block of the ion (per particle) in units if
                            the block given by `el`.
    '''
    # if there is some ion table specified in the config, use it as default
    iontbl = cloudy.config_ion_table(s.redshift) if iontbl is None else iontbl
    f_ion = 10. ** iontbl.interp_snap(el + ' ' + ionisation, s.gas,
                                      selfshield=selfshield,
                                      warn_outofbounds=warn_outofbounds)
    return f_ion * s.gas.get(el)


# this is not super correct: typically the element is taken from the block
# 'elements', but could in principle also be defined seperately!
calc_ion_mass._deps = set(['H', 'mass', 'rho', 'temp', 'elements'])


def get_luminosities(stars, band='bolometric', IMF=None):
    '''
    Get the luminosities for a stellar sub-snapshot for a given band.

    This function calls `ssp.inter_bc_qty` and, hence, is using

    Args:
        stars (Snap):   The stellar snapshot to interpolate the luminosities for.
        band (str):     The band in which to interpolate the luminosities for.
                        Possible choices are:
                        'bolometric', 'U', 'B', 'V', 'R', and 'K',
        IMF (str):      The name of the IMF to use, if folder and base is not
                        given (undefined behaviour otherwise). By default the one
                        given in `gadget.cfg` is used. Values that are not in
                        `table_base_name.keys()` are invalid.

    Returns:
        lum (UnitArr):  The luminosities of the star particles.
    '''
    from ..ssp import inter_bc_qty
    from ..physics import solar

    if band == 'bolometric':
        qty = 'Mbol'
    else:
        qty = '%smag' % band.upper()

    mag_per_Msol = inter_bc_qty(stars['age'], stars['metallicity'], qty=qty,
                                units='mag', IMF=IMF)
    lum_per_Msol = UnitQty(10. ** (-0.4 * (mag_per_Msol - solar.abs_mag)), 'Lsol')
    return (stars['mass'] / UnitScalar('1 Msol')).in_units_of(1, subs=stars) * lum_per_Msol


get_luminosities._deps = set(['age', 'metallicity', 'mass'])




'''
Mapping quantities onto a 2D-grid.

Example:
    >>> from ..analysis import *
    >>> from ..transformation import *
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=True)
    >>> Translation(UnitArr([-48087.1,-49337.1,-46084.3],'kpc')).apply(s)
    >>> s['vel'] -= UnitArr([-42.75,-15.60,-112.20],'km s**-1')
    load block vel... done.
    >>> orientate_at(s[s['r'] < '10 kpc'].baryons, 'L', total=True)
    load block pos... done.
    apply stored Translation to block pos... done.
    derive block r... done.
    load block mass... done.
    derive block momentum... done.
    derive block angmom... done.
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    apply Rotation to "vel" of "snap_M1196_4x_470"... done.
    >>> sub = s[BoxMask('120 kpc', sph_overlap=True)]
    load block hsml... done.
    >>> m_b = map_qty(sub.baryons, '120 kpc', False, 'mass', Npx=256)
    create a 256 x 256 map (120 x 120 [kpc])...
    load block rho... done.
    derive block dV... done.
    create a 256 x 256 SPH-grid (120 x 120 [kpc])...
    done with SPH grid
    >>> m_b # doctest: +ELLIPSIS
    <Map at 0x...; units="Msol", Npx=(256, 256)>
    >>> m_s = map_qty(sub.stars, '120 kpc', False, 'mass', Npx=256)
    create a 256 x 256 map (120 x 120 [kpc])...
    >>> m_g = map_qty(sub.gas, '120 kpc', False, 'mass', Npx=256)
    create a 256 x 256 map (120 x 120 [kpc])...
    create a 256 x 256 SPH-grid (120 x 120 [kpc])...
    done with SPH grid
    >>> rel_err = np.abs(m_b - (m_s+m_g)) / m_b
    >>> if rel_err.max() > 1e-10:
    ...     print(np.percentile(rel_err, [50,90,95,100]))
'''
__all__ = ['map_qty']

from ..units import *
from ..snapshot import *
from .core import *
from ..gadget import *
from ..kernels import *
from .. import environment
import numpy as np

def map_qty(s, extent, field, qty, av=None, reduction=None, Npx=256,
            xaxis=0, yaxis=1, softening=None, sph=True, kernel=None, dV='dV'):
    '''
    A fast routine for binning SPH quantities onto a map.

    Args:
        s (Snap):           The (sub-)snapshot to bin from.
        extent (UnitQty):   The extent of the map. It can be a scalar and then is
                            taken to be the total side length of a square around
                            the origin or a sequence of the minima and maxima of
                            the directions: [[xmin,xmax],[ymin,ymax]]
        field (bool):       If no `reduction` is given, this determines whether
                            the SPH-quantity is interpreted as a density-field or
                            its integral quantity.
                            For instance: rho would be the density-field of the
                            integral quantity mass.
        qty (UnitQty, str): The quantity to map. It can be a UnitArr of length
                            L=len(s) and dimension 1 (i.e. shape (L,)) or a string
                            that can be passed to s.get and returns such an array.
        av (UnitQty, str):  The quantity to average over. Otherwise as 'qty'.
        reduction (str):    If not None, interpret the SPH quantity not as a SPH
                            field, but as a particle property and reduce with this
                            given method along the third axis / line of sight.
                            See `SPH_to_2Dgrid_by_particle` for more information.
        Npx (int, sequence):The number of pixel per side. Either an integer that
                            is taken for both sides or a pair of such, the first
                            for the x-direction, the second for the y-direction.
        xaxis (int):        The coordinate for the x-axis. (0:'x', 1:'y', 2:'z')
        yaxis (int):        The coordinate for the y-axis. (0:'x', 1:'y', 2:'z')
        softening (UnitQty):A list of te softening lengthes that are taken for
                            smoothing the maps of the paticle types. Is
                            consequently has to have length 6. The entries for gas
                            particle types are ignored, if sph=True.
                            Default: None.
        sph (bool):         If set to True, do not use the softening length for
                            smoothing the SPH particles' quantity, but the actual
                            SPH smoothing lengthes, which differ from particle to
                            particle.
        kernel (str):       The kernel to use for smoothing. (Default: 'kernel'
                            from config file `gadget.cfg`)
        dV (UnitQty, str):  The volume elements of the particles.

    Returns:
        grid (Map):         The quantity summed along the third axis over the area
                            of a pixel (column -- *not* column density).
    '''
    zaxis = (set([0,1,2]) - set([xaxis, yaxis])).pop()
    if set([xaxis, yaxis, zaxis]) != set([0,1,2]):
        raise ValueError('Illdefined axes (x=%s, y=%s)!' % (xaxis, yaxis))

    if kernel is None:
        kernel = general['kernel']

    if isinstance(qty, str):
        qty = s.get(qty)

    if av is not None:
        if isinstance(av, str):
            av = s.get(av)
        if reduction is None:
            grid = map_qty(s, extent, field, av*qty, av=None, Npx=Npx,
                           xaxis=xaxis, yaxis=yaxis, softening=softening,
                           sph=sph, kernel=kernel, dV=dV)
            norm = map_qty(s, extent, field, av, av=None, Npx=Npx,
                           xaxis=xaxis, yaxis=yaxis, softening=softening,
                           sph=sph, kernel=kernel, dV=dV)
            grid /= norm
            grid[np.isnan(grid)] = 0.0
            return grid

    # prepare arguments
    extent, Npx, res = grid_props(extent=extent, Npx=Npx, dim=2)
    extent = extent.in_units_of(s['pos'].units, subs=s)
    res = res.in_units_of(s['pos'].units, subs=s)
    if softening is not None:
        #softening = UnitArr([0.2, 0.45, 2.52, 20.0, 0.2, 0.2],'ckpc / h_0')
        softening = UnitQty(softening, s['pos'].units, subs=s)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('create a %d x %d map (%.4g x %.4g %s)...' % (tuple(Npx) + \
                (extent[0,1]-extent[0,0],
                 extent[1,1]-extent[1,0], extent.units)))

    grid = np.zeros(Npx)
    if isinstance(qty, UnitArr):
        grid = UnitArr(grid,
                (qty.units*s['pos'].units).gather() if field else qty.units)

    if len(s) == 0:
        return Map(grid, extent=extent)

    if sph:
        sph = s.gas
        sph_qty = qty[s.gas._mask]
        if len(sph) != 0:
            if isinstance(dV, str):
                dV = s.gas.get(dV)
            dV = UnitQty(dV, sph['pos'].units**3)
            if reduction is not None:
                from .cbinning import SPH_to_2Dgrid_by_particle
                sph_binned = SPH_to_2Dgrid_by_particle(sph, qty=sph_qty,
                                                       av=av, dV=dV,
                                                       extent=extent,
                                                       Npx=Npx,
                                                       reduction=reduction,
                                                       xaxis=xaxis,
                                                       yaxis=yaxis,
                                                       kernel=kernel)
            else:
                from .cbinning import SPH_to_2Dgrid
                sph_binned = SPH_to_2Dgrid(sph,
                                           qty = sph_qty if field
                                                       else sph_qty/dV,
                                           extent=extent,
                                           Npx=Npx, xaxis=xaxis, yaxis=yaxis,
                                           kernel=kernel)
                if not field:
                    sph_binned *= sph_binned.vol_voxel()

            grid += sph_binned

    if sph:
        ptypes = list(set(range(6))-set(families['gas']))
    else:
        ptypes = list(range(6))
    proj_kernel = project_kernel(kernel)
    for pt in ptypes:
        sub = s.SubSnap([pt])
        if len(sub) == 0:
            continue

        if softening is not None:
            border = int( min(max(1, np.ceil(softening[pt] / res.min())),
                              0.3 * Npx.max()) )
            Npx_w = Npx + [2*border]*2
            extent_w = UnitArr( [extent[:,0] - border*res,
                                 extent[:,1] + border*res],
                               extent.units).T

            tmp = gridbin(sub['pos'][:,(xaxis,yaxis)], qty[sub._mask], extent=extent_w,
                          bins=Npx_w, nanval=0.0)
            tmp = smooth(tmp, softening[pt] / res.min(), kernel=proj_kernel)
            tmp = tmp[border:-border,border:-border]
        else:
            Npx_w = Npx
            extent_w = UnitArr( [extent[:,0], extent[:,1]],
                               extent.units).T
            tmp = gridbin(sub['pos'][:,(xaxis,yaxis)], qty[sub._mask], extent=extent_w,
                          bins=Npx_w, nanval=0.0)
        grid += tmp

    return Map(grid, extent=extent)




'''
A module for octrees on 3-dim. points in form of np.ndarrays.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(coctree)
    TestResults(failed=0, attempted=48)
    >>> doctest.testmod(octree)
    TestResults(failed=0, attempted=29)
'''

from .coctree import *
from .octree import *




"""
Basic loading of the fast C library.
"""
from ctypes import (
    POINTER,
    byref,
    c_char_p,
    c_double,
    c_int,
    c_size_t,
    c_uint,
    c_void_p,
    cdll,
    create_string_buffer,
)
from glob import glob

from .. import environment

try:
    cpygad = cdll.LoadLibrary(glob(environment.module_dir + "C/cpygad*.so")[0])
except:
    cpygad = cdll.LoadLibrary(glob(environment.module_dir + "../cpygad*.so")[0])

cpygad.cubic.restype = c_double
cpygad.cubic.argtypes = [c_double, c_double]
cpygad.quartic.restype = c_double
cpygad.quartic.argtypes = [c_double, c_double]
cpygad.quintic.restype = c_double
cpygad.quintic.argtypes = [c_double, c_double]
cpygad.Wendland_C2.restype = c_double
cpygad.Wendland_C2.argtypes = [c_double, c_double]
cpygad.Wendland_C4.restype = c_double
cpygad.Wendland_C4.argtypes = [c_double, c_double]
cpygad.Wendland_C6.restype = c_double
cpygad.Wendland_C6.argtypes = [c_double, c_double]

cpygad.Voigt.restype = c_double
cpygad.Voigt.argtypes = [c_double, c_double, c_double]



'''
Some (convenience) functions for binning.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(core)
    TestResults(failed=0, attempted=32)
    >>> doctest.testmod(cbinning)
    TestResults(failed=0, attempted=34)
    >>> doctest.testmod(mapping)
    TestResults(failed=0, attempted=13)
    >>> doctest.testmod(oneDbinning)
    TestResults(failed=0, attempted=10)
'''

from .core import *
from .cbinning import *
from .mapping import *
from .oneDbinning import *




'''
A module for controlled evaluation of Python expression.

By default the evaluator does not allow to access any underscore identifiers that
can potentially be hazardous to the system (compare e.g.
'eval("__import__('os').system('rm -rf /*')", {})' -- DO NOT TRY THIS!). In this
way the semi-private underscore attributes are actually protected at the same
time.

The evaluation of expression is also better controllable and debuggable with this
module as one can follow the evaluation process.

The drawback is, that the evaluation is slightly slower than the one of the
built-in 'eval' (by a few microseconds per call on my machine and a bit more for
the convenience 'eval').

Examples:
    >>> eval('7+4')
    11
    >>> eval('A/B*log10(12**3)/pi', {'B':1.2}, A=1.234)
    1.0597408738693963
    >>> eval('__import__("os")')
    Traceback (most recent call last):
    ...
    pygad.utils.safe_eval.EvalError: "Undefined symbol: '__import__'"
    >>> eval('arange(10).__class__')
    Traceback (most recent call last):
    ...
    pygad.utils.safe_eval.EvalError: "accessing underscore attributes is not allowed!: 'arange(10).__class__'"
    >>> e = Evaluator({'c':42}, my_math=np)
    >>> e.eval('pi*c')
    131.94689145077132
    >>> e.eval('exp(arr**2)', {'arr':np.arange(6)})
    array([1.00000000e+00, 2.71828183e+00, 5.45981500e+01, 8.10308393e+03,
           8.88611052e+06, 7.20048993e+10])
    >>> for name in iter_idents_in_expr('test.attr + more'):
    ...     print(name)
    attr
    more
    test
    >>> e.eval('a%2 == 0', {'a':np.arange(6)})
    array([ True, False,  True, False,  True, False])
    >>> e.eval('(a%2 == 0) | ((a%3 == 0))', {'a':np.arange(6)})
    array([ True, False,  True,  True,  True, False])
'''
__all__ = ['iter_idents_in_expr', 'EvalError', 'Evaluator', 'eval']

import math
import numpy as np
import operator as op
import ast
import re
import numbers
from fractions import Fraction

def iter_idents_in_expr(expr, retel=False):
    '''
    Iterate all identifiers within the Python expression.

    Args:
        expr (str):     A Python expression.
        retel (bool):   Also yield the AST element itself.

    Yields:
        ident (str):    The identifier.
        el (ast-type):  The AST element (ast.Name or ast.Attribute).
    '''
    for el in ast.walk(ast.parse(expr)):
        name = None
        if isinstance(el, ast.Name):
            name = el.id
        elif isinstance(el, ast.Attribute):
            name = el.attr
        if name:
            if retel:
                yield name, el
            else:
                yield name

class EvalError(Exception):
    '''
    Exception class for any errors that arise from the evaluator class.

    Args:
        msg (object):   The message to be displayed.
    '''
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)

class Evaluator(object):
    '''
    A class to evaluation expressions in a controlled way.

    See module documentation for more information.

    Args:
        namespace (dict):   Additional constants (numbers only) and callables to
                            use.
        my_math (module):   The module to take the standard functions and
                            constants from. The standard choices are math and
                            numpy. (default: numpy)
        truediv (bool):     Whether to use true division (1/2 -> 0.5).
        allow_underscore (bool):
                            Whether to allow to access underscore names.
    '''

    def __init__(self, namespace=None, my_math=None, truediv=True,
                 allow_underscore=False):
        if my_math is None:
            my_math = np

        self.allow_underscore = allow_underscore

        self.bin_op = {
                ast.Add: op.add,
                ast.Sub: op.sub,
                ast.Mult: op.mul,
                ast.Div: op.truediv if truediv else op.div,
                ast.Mod: op.mod,
                ast.Pow: op.pow,
                ast.Eq: op.eq,
                ast.Gt: op.gt,
                ast.GtE: op.ge,
                ast.Lt: op.lt,
                ast.LtE: op.le,
                ast.Is: op.is_,
                ast.IsNot: op.is_not,
                ast.BitAnd: np.bitwise_and,
                ast.BitOr: np.bitwise_or,
                ast.BitXor: np.bitwise_xor,
                }
        self.un_op = {
                ast.UAdd: op.pos,
                ast.USub: op.neg,
                ast.Not: op.not_,
                }

        self.namespace = {'None':None, 'True':True, 'False':False,
                'Fraction':Fraction}
        for ns in [my_math.__dict__, namespace if namespace is not None else {}]:
            self.namespace.update( { name:val for name,val in ns.items()
                    if hasattr(val,'__call__')
                    or isinstance(val,(numbers.Number,str,np.ndarray)) } )

    def _eval(self, node):
        if hasattr(node,'ctx') and not isinstance(node.ctx, ast.Load):
            raise EvalError('only loading context allowed')
        if isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Name):
            return self.variables[node.id]
        elif isinstance(node, ast.BinOp):
            return self.bin_op[type(node.op)](self._eval(node.left), self._eval(node.right))
        elif isinstance(node, ast.UnaryOp):
            return self.un_op[type(node.op)](self._eval(node.operand))
        elif isinstance(node, ast.Call):
#           the following lines are removed: incompatible since Pyhton 3.5
#           if node.kwargs:
#               raise NotImplementedError('kwargs are not supported')
#           if node.starargs:
#               raise NotImplementedError('starargs are not supported')
            kwargs = { kw.arg: self._eval(kw.value) for kw in node.keywords }
            posargs = [self._eval(arg) for arg in node.args]
            return self._eval(node.func)(*posargs,**kwargs)
        elif isinstance(node, ast.Attribute):
            if node.attr.startswith('_') and not self.allow_underscore:
                raise EvalError('accessing underscore attributes is not allowed!')
            return getattr(self._eval(node.value), node.attr)
        elif isinstance(node, ast.IfExp):
            if self._eval(node.test):
                return self._eval(node.body)
            else:
                return self._eval(node.orelse)
        elif isinstance(node, ast.Compare):
            left = self._eval(node.left)
            if len(node.ops) == 1:
                # might be an array-like comparison, that does not work in the
                # general case of multiple comparisons at once
                operator, comparator = node.ops[0], node.comparators[0]
                right = self._eval(comparator)
                return self.bin_op[type(operator)](left, right)
            for operator,comparator in zip(node.ops,node.comparators):
                right = self._eval(comparator)
                if not self.bin_op[type(operator)](left, right):
                    return False
                left = right
            return True
        elif isinstance(node, ast.BoolOp):
            for val in node.values:
                val = self._eval(val)
                if isinstance(node.op, ast.Or) and val:
                    return True
                if isinstance(node.op, ast.And) and not val:
                    return False
            return isinstance(node.op, ast.And)
        elif isinstance(node, ast.Subscript):
            return self._eval(node.value)[self._eval(node.slice)]
        elif isinstance(node, ast.Slice):
            start = self._eval(node.lower)
            stop  = self._eval(node.upper)
            step  = self._eval(node.step)
            return slice(start, stop, step)
        elif isinstance(node, ast.ExtSlice):
            return tuple(map(self._eval, node.dims))
        elif isinstance(node, ast.Index):
            return self._eval(node.value)
        elif isinstance(node, ast.Str):
            return node.s
        elif node is None:
            return None
        elif isinstance(node, ast.List):
            return list(map(self._eval, node.elts))
        elif isinstance(node, ast.Tuple):
            return tuple(map(self._eval, node.elts))
        elif isinstance(node, ast.NameConstant):
            return node.value
        else:
            # seldomly reached due to substitution of spaces
            raise EvalError('AST node type %s not supported' %
                                node.__class__.__name__)

    def eval(self, expr, idents=None):
        '''
        Evaluate the expression.

        Args:
            expr (str):     The expression with Python syntax to evaluate.
            idents (dict):  Additional identifiers to use during the evalutations.

        Returns:
            value (object): The value of the expression.

        Raises:
            EvalError:      If some used Python syntax is not supported by this
                            Evaluator or of there was an unknown value in the
                            expression.
            others:         Might occur during the evaluation, e.g. OverflowError.
        '''
        self.variables = self.namespace.copy()
        if idents is not None:
            self.variables.update(idents)
        try:
            return self._eval(ast.parse(expr,mode='eval').body)
        except KeyError as e:
            if re.match(r'<class .*>', str(e)):
                raise EvalError('Undefined operation: %s' % e)
            else:
                raise EvalError('Undefined symbol: %s' % e)
        except (EvalError, NotImplementedError) as e:
            raise EvalError("%s: '%s'" % (str(e).strip("'"), expr))
        except:
            raise

def eval(expr, namespace=None, my_math=None, truediv=True,
         allow_underscore=False, **kwargs):
    '''
    Evaluate an expression.

    This is a convenience function to call Evaluator().eval. Hence, for more
    details see Evaluator and Evaluator.eval.

    Args:
        expr (str):         The expression to evaluate.
        namespace (dict):   The namespace to operate on.
        my_math (module):   The module to take the standard functions and
                            constants from. The standard choices are math and
                            numpy. (default: numpy)
        truediv (bool):     Whether to use true division (1/2 -> 0.5).
        allow_underscore (bool):
                            Whether to allow to access underscore names.
        **kwargs:           A convenient way to set varibales in the evaluation
                            namespace.

    Returns:
        value (object):     The evaluated expression.

    Raises:
        EvalError:      If some used Python syntax is not supported by this
                        Evaluator or of there was an unknown value in the
                        expression.
        others:         Might occur during the evaluation, e.g. OverflowError.

    Example:
        >>> eval('2*pi*R', R=2.0)
        12.566370614359172
    '''
    e = Evaluator(namespace=namespace, my_math=my_math, truediv=truediv,
                  allow_underscore=allow_underscore)
    return e.eval(expr, idents=kwargs)
